import os
from typing import TYPE_CHECKING

import pytest
from syrupy.extensions.amber import AmberSnapshotExtension

if TYPE_CHECKING:
    from syrupy.location import PyTestLocation
    from syrupy.types import SnapshotIndex


@pytest.fixture(scope="session", autouse=True)
def unset_dagster_home():
    old_env = os.getenv("DAGSTER_HOME")
    if old_env is not None:
        del os.environ["DAGSTER_HOME"]
    yield
    if old_env is not None:
        os.environ["DAGSTER_HOME"] = old_env


class SharedSnapshotExtension(AmberSnapshotExtension):
    @classmethod
    def get_snapshot_name(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls,
        *,
        index: "SnapshotIndex",
        test_location: "PyTestLocation",
    ) -> str:
        snapshot_name = test_location.snapshot_name

        # Exclude any of the GraphQLContextVariant suffixes from the snapshot name
        # so that we don't have to re-generate an identical snapshot for each one
        variant_start_index = snapshot_name.find("[")
        if variant_start_index != -1:
            snapshot_name = snapshot_name[:variant_start_index]

        return f"{snapshot_name}[{index}]"


@pytest.fixture
def snapshot(snapshot):
    return snapshot.use_extension(SharedSnapshotExtension)



from time import sleep

from dagster import Config, In, Int, List, Out, Output, job, op
from pydantic import Field


@op(
    ins={"units": In(List[Int])},
    out={"total": Out(Int)},
)
def sleeper(context, units):
    tot = 0
    for sec in units:
        context.log.info(f"Sleeping for {sec} seconds")
        sleep(sec)
        tot += sec

    return tot


class GiverConfig(Config):
    units: list[int] = Field(default_value=[1, 1, 1, 1])  # type: ignore


@op(
    out={
        "out 1": Out(List[Int]),
        "out 2": Out(List[Int]),
        "out 3": Out(List[Int]),
        "out 4": Out(List[Int]),
    },
)
def giver(config: GiverConfig):
    queues = [[], [], [], []]
    for i, sec in enumerate(config.units):
        queues[i % 4].append(sec)

    yield Output(queues[0], "out_1")
    yield Output(queues[1], "out_2")
    yield Output(queues[2], "out_3")
    yield Output(queues[3], "out_4")


@op(
    ins={
        "in_1": In(Int),
        "in_2": In(Int),
        "in_3": In(Int),
        "in_4": In(Int),
    },
    out=Out(int),
)
def total(_, in_1, in_2, in_3, in_4):
    return in_1 + in_2 + in_3 + in_4


@job(
    description="Demo diamond-shaped job that has four-path parallel structure of ops.",
)
def sleepy_job():
    giver_res = giver()

    total(
        in_1=sleeper(units=giver_res.out_1),
        in_2=sleeper(units=giver_res.out_2),
        in_3=sleeper(units=giver_res.out_3),
        in_4=sleeper(units=giver_res.out_4),
    )



import time
from collections.abc import Sequence
from typing import TYPE_CHECKING, Optional

import dagster._check as check
from dagster._core.definitions.run_request import InstigatorType
from dagster._core.definitions.selector import JobSubsetSelector, RepositorySelector, SensorSelector
from dagster._core.scheduler.instigation import (
    InstigatorState,
    InstigatorStatus,
    SensorInstigatorData,
)
from dagster._core.workspace.permissions import Permissions

from dagster_graphql.implementation.loader import RepositoryScopedBatchLoader
from dagster_graphql.implementation.utils import (
    UserFacingGraphQLError,
    assert_permission,
    assert_permission_for_location,
)
from dagster_graphql.schema.util import ResolveInfo

if TYPE_CHECKING:
    from dagster_graphql.schema.instigation import GrapheneDryRunInstigationTick
    from dagster_graphql.schema.sensors import (
        GrapheneSensor,
        GrapheneSensors,
        GrapheneStopSensorMutationResult,
    )


def get_sensors_or_error(
    graphene_info: ResolveInfo,
    repository_selector: RepositorySelector,
    instigator_statuses: set[InstigatorStatus] | None = None,
) -> "GrapheneSensors":
    from dagster_graphql.schema.sensors import GrapheneSensor, GrapheneSensors

    check.inst_param(repository_selector, "repository_selector", RepositorySelector)

    location = graphene_info.context.get_code_location(repository_selector.location_name)
    repository = location.get_repository(repository_selector.repository_name)
    batch_loader = RepositoryScopedBatchLoader(graphene_info.context.instance, repository)
    sensors = repository.get_sensors()
    sensor_states = graphene_info.context.instance.all_instigator_state(
        repository_origin_id=repository.get_remote_origin_id(),
        repository_selector_id=repository_selector.selector_id,
        instigator_type=InstigatorType.SENSOR,
        instigator_statuses=instigator_statuses,
    )

    sensor_states_by_name = {state.name: state for state in sensor_states}
    if instigator_statuses:
        filtered = [
            sensor
            for sensor in sensors
            if sensor.get_current_instigator_state(sensor_states_by_name.get(sensor.name)).status
            in instigator_statuses
        ]
    else:
        filtered = sensors

    return GrapheneSensors(
        results=[
            GrapheneSensor(
                sensor,
                sensor_states_by_name.get(sensor.name),
                batch_loader,
            )
            for sensor in filtered
        ]
    )


def get_sensor_or_error(graphene_info: ResolveInfo, selector: SensorSelector) -> "GrapheneSensor":
    from dagster_graphql.schema.errors import GrapheneSensorNotFoundError
    from dagster_graphql.schema.sensors import GrapheneSensor

    check.inst_param(selector, "selector", SensorSelector)

    sensor = graphene_info.context.get_sensor(selector)
    if not sensor:
        raise UserFacingGraphQLError(GrapheneSensorNotFoundError(selector.sensor_name))

    sensor_state = graphene_info.context.instance.get_instigator_state(
        sensor.get_remote_origin_id(),
        sensor.selector_id,
    )

    return GrapheneSensor(sensor, sensor_state)


def start_sensor(graphene_info: ResolveInfo, sensor_selector: SensorSelector) -> "GrapheneSensor":
    from dagster_graphql.schema.errors import GrapheneSensorNotFoundError
    from dagster_graphql.schema.sensors import GrapheneSensor

    check.inst_param(sensor_selector, "sensor_selector", SensorSelector)

    sensor = graphene_info.context.get_sensor(sensor_selector)
    if not sensor:
        raise UserFacingGraphQLError(GrapheneSensorNotFoundError(sensor_selector.sensor_name))

    sensor_state = graphene_info.context.instance.start_sensor(sensor)
    return GrapheneSensor(sensor, sensor_state)


def stop_sensor(
    graphene_info: ResolveInfo, instigator_origin_id: str, instigator_selector_id: str
) -> "GrapheneStopSensorMutationResult":
    from dagster_graphql.schema.sensors import GrapheneStopSensorMutationResult

    check.str_param(instigator_origin_id, "instigator_origin_id")
    instance = graphene_info.context.instance

    sensors = {
        sensor.get_remote_origin_id(): sensor
        for code_location in graphene_info.context.code_locations
        for repository in code_location.get_repositories().values()
        for sensor in repository.get_sensors()
    }

    sensor = sensors.get(instigator_origin_id)
    if sensor:
        assert_permission_for_location(
            graphene_info,
            Permissions.EDIT_SENSOR,
            sensor.selector.location_name,
        )
    else:
        assert_permission(
            graphene_info,
            Permissions.EDIT_SENSOR,
        )

    state = instance.stop_sensor(
        instigator_origin_id,
        instigator_selector_id,
        sensor,
    )
    return GrapheneStopSensorMutationResult(state)


def reset_sensor(graphene_info: ResolveInfo, sensor_selector: SensorSelector) -> "GrapheneSensor":
    from dagster_graphql.schema.errors import GrapheneSensorNotFoundError
    from dagster_graphql.schema.sensors import GrapheneSensor

    check.inst_param(sensor_selector, "sensor_selector", SensorSelector)
    sensor = graphene_info.context.get_sensor(sensor_selector)
    if not sensor:
        raise UserFacingGraphQLError(GrapheneSensorNotFoundError(sensor_selector.sensor_name))

    sensor_state = graphene_info.context.instance.reset_sensor(sensor)

    return GrapheneSensor(sensor, sensor_state)


def get_sensors_for_job(
    graphene_info: ResolveInfo, selector: JobSubsetSelector
) -> Sequence["GrapheneSensor"]:
    from dagster_graphql.schema.sensors import GrapheneSensor

    check.inst_param(selector, "selector", JobSubsetSelector)

    sensors = graphene_info.context.get_sensors_targeting_job(selector)

    results = []
    for sensor in sensors:
        sensor_state = graphene_info.context.instance.get_instigator_state(
            sensor.get_remote_origin_id(),
            sensor.selector_id,
        )
        results.append(GrapheneSensor(sensor, sensor_state))

    return results


def get_sensor_next_tick(
    graphene_info: ResolveInfo, sensor_state: InstigatorState
) -> Optional["GrapheneDryRunInstigationTick"]:
    from dagster_graphql.schema.instigation import GrapheneDryRunInstigationTick

    check.inst_param(sensor_state, "sensor_state", InstigatorState)
    if not sensor_state.is_running:
        return None

    repository_origin = sensor_state.origin.repository_origin
    selector = SensorSelector(
        location_name=repository_origin.code_location_origin.location_name,
        repository_name=repository_origin.repository_name,
        sensor_name=sensor_state.name,
    )

    sensor = graphene_info.context.get_sensor(selector)
    if not sensor:
        return None

    ticks = graphene_info.context.instance.get_ticks(
        sensor_state.instigator_origin_id, sensor_state.selector_id, limit=1
    )
    if not ticks:
        return None
    latest_tick = ticks[0]

    next_timestamp = latest_tick.timestamp + sensor.min_interval_seconds
    if next_timestamp < time.time():
        return None
    return GrapheneDryRunInstigationTick(sensor.sensor_selector, next_timestamp)


def set_sensor_cursor(
    graphene_info: ResolveInfo, selector: SensorSelector, cursor: str | None
) -> "GrapheneSensor":
    check.inst_param(selector, "selector", SensorSelector)
    check.opt_str_param(cursor, "cursor")

    from dagster_graphql.schema.errors import GrapheneSensorNotFoundError
    from dagster_graphql.schema.sensors import GrapheneSensor

    sensor = graphene_info.context.get_sensor(selector)
    if not sensor:
        raise UserFacingGraphQLError(GrapheneSensorNotFoundError(selector.sensor_name))

    instance = graphene_info.context.instance
    stored_state = instance.get_instigator_state(
        sensor.get_remote_origin_id(),
        sensor.selector_id,
    )
    sensor_state = sensor.get_current_instigator_state(stored_state)
    instigator_data = sensor_state.instigator_data
    if not isinstance(instigator_data, SensorInstigatorData):
        check.failed("Expected SensorInstigatorData")
    updated_state = sensor_state.with_data(
        SensorInstigatorData(
            last_tick_timestamp=instigator_data.last_tick_timestamp,
            last_run_key=instigator_data.last_run_key,
            min_interval=sensor.min_interval_seconds,
            cursor=cursor,
            sensor_type=sensor.sensor_type,
        )
    )
    if not stored_state:
        instance.add_instigator_state(updated_state)
    else:
        instance.update_instigator_state(updated_state)

    return GrapheneSensor(sensor, updated_state)



from dagster._core.workspace.context import WorkspaceRequestContext
from dagster_graphql.test.utils import infer_job_selector

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
)
from dagster_graphql_tests.graphql.repo import LONG_INT
from dagster_graphql_tests.graphql.utils import sync_execute_get_events


class TestMaterializations(ExecutingGraphQLContextTestMatrix):
    def test_materializations(self, graphql_context: WorkspaceRequestContext, snapshot):
        selector = infer_job_selector(graphql_context, "materialization_job")
        logs = sync_execute_get_events(
            context=graphql_context,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                }
            },
        )

        materializations = [log for log in logs if log["__typename"] == "MaterializationEvent"]
        assert len(materializations) == 1
        mat = materializations[0]
        assert mat["label"] == "all_types"

        entry = mat["metadataEntries"][0]
        assert entry["__typename"] == "TextMetadataEntry"
        assert entry["text"]

        entry = mat["metadataEntries"][1]
        assert entry["__typename"] == "UrlMetadataEntry"
        assert entry["url"]

        entry = mat["metadataEntries"][2]
        assert entry["__typename"] == "PathMetadataEntry"
        assert entry["path"]

        entry = mat["metadataEntries"][3]
        assert entry["__typename"] == "JsonMetadataEntry"
        assert entry["jsonString"]

        entry = mat["metadataEntries"][4]
        assert entry["__typename"] == "PythonArtifactMetadataEntry"
        assert entry["module"]
        assert entry["name"]

        entry = mat["metadataEntries"][5]
        assert entry["__typename"] == "PythonArtifactMetadataEntry"
        assert entry["module"]
        assert entry["name"]

        entry = mat["metadataEntries"][6]
        assert entry["__typename"] == "FloatMetadataEntry"
        assert entry["floatValue"]

        entry = mat["metadataEntries"][7]
        assert entry["__typename"] == "IntMetadataEntry"
        assert entry["intRepr"]

        entry = mat["metadataEntries"][8]
        assert entry["__typename"] == "FloatMetadataEntry"
        assert entry["floatValue"] is None  # float NaN test

        entry = mat["metadataEntries"][9]
        assert entry["__typename"] == "IntMetadataEntry"
        assert int(entry["intRepr"]) == LONG_INT

        entry = mat["metadataEntries"][10]
        assert entry["__typename"] == "PipelineRunMetadataEntry"
        assert entry["runId"] == "fake_run_id"

        entry = mat["metadataEntries"][11]
        assert entry["__typename"] == "AssetMetadataEntry"
        assert entry["assetKey"]
        assert entry["assetKey"]["path"]

        entry = mat["metadataEntries"][12]
        assert entry["__typename"] == "TableMetadataEntry"
        assert entry["table"]
        assert entry["table"]["records"]
        assert entry["table"]["schema"]

        entry = mat["metadataEntries"][13]
        assert entry["__typename"] == "TableSchemaMetadataEntry"
        assert entry["schema"]
        assert entry["schema"]["columns"]
        assert entry["schema"]["columns"][0]["constraints"]
        assert entry["schema"]["constraints"]
        assert entry["schema"]["columns"]
        assert entry["schema"]["columns"][0]["tags"]

        entry = mat["metadataEntries"][14]
        assert entry["__typename"] == "JobMetadataEntry"
        assert entry["jobName"]

        entry = mat["metadataEntries"][15]
        assert entry["__typename"] == "TextMetadataEntry"
        assert entry["text"] == "SomeClass"

        entry = mat["metadataEntries"][16]
        assert entry["__typename"] == "FloatMetadataEntry"
        assert entry["floatValue"] is None
        assert entry["floatRepr"] == "inf"

        entry = mat["metadataEntries"][17]
        assert entry["__typename"] == "FloatMetadataEntry"
        assert entry["floatValue"] is None
        assert entry["floatRepr"] == "-inf"

        non_engine_event_logs = [
            message for message in logs if message["__typename"] != "EngineEvent"
        ]

        snapshot.assert_match([message["__typename"] for message in non_engine_event_logs])



from unittest import mock

from dagster_graphql.test.utils import execute_dagster_graphql

SET_NUX_SEEN_MUTATION = """
    mutation SetNuxSeen {
       setNuxSeen
    }
"""

GET_SHOULD_SHOW_NUX_QUERY = """
  query ShouldShowNux {
    shouldShowNux
  }

"""


def test_stores_nux_seen_state(graphql_context):
    result = execute_dagster_graphql(graphql_context, GET_SHOULD_SHOW_NUX_QUERY)
    assert not result.errors
    assert result.data
    assert result.data["shouldShowNux"] is True

    execute_dagster_graphql(graphql_context, SET_NUX_SEEN_MUTATION)

    result = execute_dagster_graphql(graphql_context, GET_SHOULD_SHOW_NUX_QUERY)
    assert not result.errors
    assert result.data
    assert result.data["shouldShowNux"] is False


def test_does_not_show_nux_if_read_only_filesystem(graphql_context):
    with mock.patch(
        "dagster._core.nux.nux_seen_filepath", side_effect=OSError("Read-only filesystem")
    ):
        result = execute_dagster_graphql(graphql_context, GET_SHOULD_SHOW_NUX_QUERY)
        assert not result.errors
        assert result.data
        assert result.data["shouldShowNux"] is False



from collections.abc import Sequence
from typing import TYPE_CHECKING

import dagster._check as check
from dagster import AssetKey
from dagster._core.scheduler.instigation import AutoMaterializeAssetEvaluationRecord

from dagster_graphql.schema.auto_materialize_asset_evaluations import (
    GrapheneAutoMaterializeAssetEvaluationNeedsMigrationError,
    GrapheneAutoMaterializeAssetEvaluationRecord,
    GrapheneAutoMaterializeAssetEvaluationRecords,
)

if TYPE_CHECKING:
    from dagster_graphql.schema.util import ResolveInfo


def _get_migration_error(
    graphene_info: "ResolveInfo",
) -> GrapheneAutoMaterializeAssetEvaluationNeedsMigrationError | None:
    if graphene_info.context.instance.schedule_storage is None:
        return GrapheneAutoMaterializeAssetEvaluationNeedsMigrationError(
            message="Instance does not have schedule storage configured, cannot fetch evaluations."
        )
    if not graphene_info.context.instance.schedule_storage.supports_auto_materialize_asset_evaluations:
        return GrapheneAutoMaterializeAssetEvaluationNeedsMigrationError(
            message=(
                "Auto materialize evaluations are not getting logged. Run `dagster instance"
                " migrate` to enable."
            )
        )
    return None


def _get_graphene_records_from_evaluations(
    graphene_info: "ResolveInfo", evaluation_records: Sequence[AutoMaterializeAssetEvaluationRecord]
) -> GrapheneAutoMaterializeAssetEvaluationRecords:
    return GrapheneAutoMaterializeAssetEvaluationRecords(
        records=[
            GrapheneAutoMaterializeAssetEvaluationRecord(evaluation)
            for evaluation in evaluation_records
        ]
    )


def fetch_auto_materialize_asset_evaluations(
    graphene_info: "ResolveInfo",
    asset_key: AssetKey,
    limit: int,
    cursor: str | None,
):
    """Fetch asset policy evaluations from storage."""
    migration_error = _get_migration_error(graphene_info)
    if migration_error:
        return migration_error

    schedule_storage = check.not_none(graphene_info.context.instance.schedule_storage)
    return _get_graphene_records_from_evaluations(
        graphene_info,
        schedule_storage.get_auto_materialize_asset_evaluations(
            key=asset_key,
            limit=limit,
            cursor=int(cursor) if cursor else None,
        ),
    )


def fetch_auto_materialize_asset_evaluations_for_evaluation_id(
    graphene_info: "ResolveInfo",
    evaluation_id: int,
):
    migration_error = _get_migration_error(graphene_info)
    if migration_error:
        return migration_error

    schedule_storage = check.not_none(graphene_info.context.instance.schedule_storage)

    return _get_graphene_records_from_evaluations(
        graphene_info,
        schedule_storage.get_auto_materialize_evaluations_for_evaluation_id(
            evaluation_id=evaluation_id,
        ),
    )



import os
from time import sleep

from dagster._core.events import RunFailureReason
from dagster._core.execution.plan.resume_retry import ReexecutionStrategy
from dagster._core.storage.dagster_run import DagsterRunStatus
from dagster._core.storage.tags import RESUME_RETRY_TAG, RUN_FAILURE_REASON_TAG
from dagster._core.test_utils import create_run_for_test, poll_for_finished_run
from dagster._core.workspace.context import WorkspaceRequestContext
from dagster_graphql.client.query import (
    LAUNCH_PIPELINE_EXECUTION_MUTATION,
    LAUNCH_PIPELINE_REEXECUTION_MUTATION,
    PIPELINE_REEXECUTION_INFO_QUERY,
)
from dagster_graphql.schema.inputs import GrapheneReexecutionStrategy
from dagster_graphql.test.utils import (
    execute_dagster_graphql,
    execute_dagster_graphql_and_finish_runs,
    infer_job_selector,
    main_repo_location_name,
)
from dagster_shared.seven.temp_dir import get_system_temp_directory

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
    ReadonlyGraphQLContextTestMatrix,
)
from dagster_graphql_tests.graphql.repo import (
    csv_hello_world_ops_config,
    get_retry_multi_execution_params,
    retry_config,
)
from dagster_graphql_tests.graphql.utils import (
    get_all_logs_for_finished_run_via_subscription,
    step_did_fail,
    step_did_fail_in_records,
    step_did_not_run,
    step_did_not_run_in_records,
    step_did_skip,
    step_did_succeed,
    step_did_succeed_in_records,
    step_started,
)


def first_event_of_type(logs, message_type):
    for log in logs:
        if log["__typename"] == message_type:
            return log
    return None


def has_event_of_type(logs, message_type):
    return first_event_of_type(logs, message_type) is not None


def get_step_output_event(logs, step_key, output_name="result"):
    for log in logs:
        if (
            log["__typename"] == "ExecutionStepOutputEvent"
            and log["stepKey"] == step_key
            and log["outputName"] == output_name
        ):
            return log

    return None


class TestRetryExecutionReadonly(ReadonlyGraphQLContextTestMatrix):
    def test_retry_execution_permission_failure(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "eventually_successful")

        code_location = graphql_context.get_code_location(main_repo_location_name())
        repository = code_location.get_repository("test_repo")
        remote_job_origin = repository.get_full_job("eventually_successful").get_remote_origin()

        run_id = create_run_for_test(
            graphql_context.instance,
            "eventually_successful",
            remote_job_origin=remote_job_origin,
        ).run_id

        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": {},
                    "executionMetadata": {
                        "rootRunId": run_id,
                        "parentRunId": run_id,
                        "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                    },
                }
            },
        )
        assert result.data["launchPipelineReexecution"]["__typename"] == "UnauthorizedError"


class TestRetryExecution(ExecutingGraphQLContextTestMatrix):
    def test_retry_pipeline_execution(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "eventually_successful")
        result = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": retry_config(0),
                }
            },
        )

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            "pipelineRunLogs"
        ]["messages"]

        assert step_did_succeed(logs, "spawn")
        assert step_did_fail(logs, "fail")
        assert step_did_not_run(logs, "fail_2")
        assert step_did_not_run(logs, "fail_3")
        assert step_did_not_run(logs, "reset")
        assert step_did_not_run(logs, "collect")

        retry_one = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": retry_config(1),
                    "executionMetadata": {
                        "rootRunId": run_id,
                        "parentRunId": run_id,
                        "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                    },
                }
            },
        )

        run_id = retry_one.data["launchPipelineReexecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            "pipelineRunLogs"
        ]["messages"]
        assert step_did_not_run(logs, "spawn")
        assert step_did_succeed(logs, "fail")
        assert step_did_fail(logs, "fail_2")
        assert step_did_not_run(logs, "fail_3")
        assert step_did_not_run(logs, "reset")
        assert step_did_not_run(logs, "collect")

        retry_two = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": retry_config(2),
                    "executionMetadata": {
                        "rootRunId": run_id,
                        "parentRunId": run_id,
                        "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                    },
                }
            },
        )

        run_id = retry_two.data["launchPipelineReexecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            "pipelineRunLogs"
        ]["messages"]

        assert step_did_not_run(logs, "spawn")
        assert step_did_not_run(logs, "fail")
        assert step_did_succeed(logs, "fail_2")
        assert step_did_fail(logs, "fail_3")
        assert step_did_not_run(logs, "reset")
        assert step_did_not_run(logs, "collect")

        retry_three = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": retry_config(3),
                    "executionMetadata": {
                        "rootRunId": run_id,
                        "parentRunId": run_id,
                        "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                    },
                }
            },
        )

        run_id = retry_three.data["launchPipelineReexecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            "pipelineRunLogs"
        ]["messages"]

        assert step_did_not_run(logs, "spawn")
        assert step_did_not_run(logs, "fail")
        assert step_did_not_run(logs, "fail_2")
        assert step_did_succeed(logs, "fail_3")
        assert step_did_succeed(logs, "reset")
        assert step_did_succeed(logs, "collect")

    def test_retry_resource_pipeline(self, graphql_context: WorkspaceRequestContext):
        context = graphql_context
        selector = infer_job_selector(graphql_context, "retry_resource_job")
        result = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                }
            },
        )

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(context, run_id)["pipelineRunLogs"][
            "messages"
        ]
        assert step_did_succeed(logs, "start")
        assert step_did_fail(logs, "will_fail")

        retry_one = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "executionMetadata": {
                        "rootRunId": run_id,
                        "parentRunId": run_id,
                        "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                    },
                }
            },
        )
        run_id = retry_one.data["launchPipelineReexecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(context, run_id)["pipelineRunLogs"][
            "messages"
        ]
        assert step_did_not_run(logs, "start")
        assert step_did_fail(logs, "will_fail")

    def test_retry_multi_output(self, graphql_context: WorkspaceRequestContext):
        context = graphql_context
        result = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": get_retry_multi_execution_params(context, should_fail=True)
            },
        )

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(context, run_id)["pipelineRunLogs"][
            "messages"
        ]
        assert step_did_succeed(logs, "multi")
        assert step_did_skip(logs, "child_multi_skip")
        assert step_did_fail(logs, "can_fail")
        assert step_did_not_run(logs, "child_fail")
        assert step_did_not_run(logs, "child_skip")
        assert step_did_not_run(logs, "grandchild_fail")

        retry_one = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": get_retry_multi_execution_params(
                    context, should_fail=True, retry_id=run_id
                )
            },
        )

        run_id = retry_one.data["launchPipelineReexecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(context, run_id)["pipelineRunLogs"][
            "messages"
        ]
        assert step_did_not_run(logs, "multi")
        assert step_did_not_run(logs, "child_multi_skip")
        assert step_did_fail(logs, "can_fail")
        assert step_did_not_run(logs, "child_fail")
        assert step_did_not_run(logs, "child_skip")
        assert step_did_not_run(logs, "grandchild_fail")

        retry_two = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": get_retry_multi_execution_params(
                    context, should_fail=False, retry_id=run_id
                )
            },
        )

        run_id = retry_two.data["launchPipelineReexecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(context, run_id)["pipelineRunLogs"][
            "messages"
        ]
        assert step_did_not_run(logs, "multi")
        assert step_did_not_run(logs, "child_multi_skip")
        assert step_did_succeed(logs, "can_fail")
        assert step_did_succeed(logs, "child_fail")
        assert step_did_skip(logs, "child_skip")
        assert step_did_succeed(logs, "grandchild_fail")

    def test_successful_pipeline_reexecution(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "csv_hello_world")
        result_one = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": csv_hello_world_ops_config(),
                }
            },
        )
        run_id = result_one.data["launchPipelineExecution"]["run"]["runId"]

        assert result_one.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        result = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)
        logs = result["pipelineRunLogs"]["messages"]
        assert get_step_output_event(logs, "sum_op")
        assert get_step_output_event(logs, "sum_sq_op")

        # retry
        result_two = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": csv_hello_world_ops_config(),
                    "stepKeys": ["sum_sq_op"],
                    "executionMetadata": {
                        "rootRunId": run_id,
                        "parentRunId": run_id,
                        "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                    },
                }
            },
        )

        query_result = result_two.data["launchPipelineReexecution"]
        assert query_result["__typename"] == "LaunchRunSuccess"

        new_run_id = query_result["run"]["runId"]
        result = get_all_logs_for_finished_run_via_subscription(graphql_context, new_run_id)
        logs = result["pipelineRunLogs"]["messages"]

        assert isinstance(logs, list)
        assert has_event_of_type(logs, "RunStartEvent")
        assert has_event_of_type(logs, "RunSuccessEvent")
        assert not has_event_of_type(logs, "RunFailureEvent")

        assert not get_step_output_event(logs, "sum_op")
        assert get_step_output_event(logs, "sum_sq_op")

    def test_pipeline_reexecution_info_query(
        self, graphql_context: WorkspaceRequestContext, snapshot
    ):
        context = graphql_context
        selector = infer_job_selector(graphql_context, "csv_hello_world")

        result = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": csv_hello_world_ops_config(),
                }
            },
        )
        run_id = result.data["launchPipelineExecution"]["run"]["runId"]

        # retry
        retry_result = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": csv_hello_world_ops_config(),
                    "stepKeys": ["sum_sq_op"],
                    "executionMetadata": {
                        "rootRunId": run_id,
                        "parentRunId": run_id,
                        "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                    },
                }
            },
        )
        new_run_id = retry_result.data["launchPipelineReexecution"]["run"]["runId"]

        info_result_one = execute_dagster_graphql_and_finish_runs(
            context, PIPELINE_REEXECUTION_INFO_QUERY, variables={"runId": run_id}
        )
        query_result_one = info_result_one.data["pipelineRunOrError"]
        assert query_result_one["__typename"] == "Run"
        assert query_result_one["stepKeysToExecute"] == ["sum_op", "sum_sq_op"]  # full execution

        info_result_two = execute_dagster_graphql_and_finish_runs(
            context, PIPELINE_REEXECUTION_INFO_QUERY, variables={"runId": new_run_id}
        )
        query_result_two = info_result_two.data["pipelineRunOrError"]
        assert query_result_two["__typename"] == "Run"
        stepKeysToExecute = query_result_two["stepKeysToExecute"]
        assert stepKeysToExecute == ["sum_sq_op"]  # selected key

    def test_pipeline_reexecution_invalid_step_in_subset(
        self, graphql_context: WorkspaceRequestContext
    ):
        selector = infer_job_selector(graphql_context, "csv_hello_world")
        result_one = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": csv_hello_world_ops_config(),
                }
            },
        )
        assert result_one.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
        run_id = result_one.data["launchPipelineExecution"]["run"]["runId"]

        # retry
        result_two = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": csv_hello_world_ops_config(),
                    "stepKeys": ["nope"],
                    "executionMetadata": {
                        "rootRunId": run_id,
                        "parentRunId": run_id,
                        "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                    },
                }
            },
        )

        query_result = result_two.data["launchPipelineReexecution"]
        assert query_result["__typename"] == "PythonError"
        assert query_result["className"] == "DagsterExecutionStepNotFoundError"
        assert "Can not build subset plan from unknown step: nope" in query_result["message"]


class TestHardFailures(ExecutingGraphQLContextTestMatrix):
    def test_retry_hard_failure(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "hard_failer")
        result = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": {"ops": {"hard_fail_or_0": {"config": {"fail": True}}}},
                }
            },
        )

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            "pipelineRunLogs"
        ]["messages"]

        assert step_started(logs, "hard_fail_or_0")
        assert step_did_not_run(logs, "hard_fail_or_0")
        assert step_did_not_run(logs, "increment")

        retry = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": {"ops": {"hard_fail_or_0": {"config": {"fail": False}}}},
                    "executionMetadata": {
                        "rootRunId": run_id,
                        "parentRunId": run_id,
                        "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                    },
                }
            },
        )

        run_id = retry.data["launchPipelineReexecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            "pipelineRunLogs"
        ]["messages"]
        assert step_did_succeed(logs, "hard_fail_or_0")
        assert step_did_succeed(logs, "increment")

    def test_retry_failure_all_steps_with_reexecution_params(
        self, graphql_context: WorkspaceRequestContext
    ):
        """Test with providng reexecutionParams rather than executionParams."""
        selector = infer_job_selector(graphql_context, "chained_failure_job")

        # trigger failure in the conditionally_fail op
        output_file = os.path.join(
            get_system_temp_directory(), "chained_failure_job_conditionally_fail"
        )
        try:
            with open(output_file, "w", encoding="utf8"):
                result = execute_dagster_graphql_and_finish_runs(
                    graphql_context,
                    LAUNCH_PIPELINE_EXECUTION_MUTATION,
                    variables={
                        "executionParams": {
                            "selector": selector,
                        }
                    },
                )
        finally:
            os.remove(output_file)

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        run = graphql_context.instance.get_run_by_id(run_id)
        assert run and run.status == DagsterRunStatus.FAILURE
        assert run.tags[RUN_FAILURE_REASON_TAG] == RunFailureReason.STEP_FAILURE.value

        retry = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={"reexecutionParams": {"parentRunId": run_id, "strategy": "ALL_STEPS"}},
        )

        assert retry.data["launchPipelineReexecution"].get("run"), retry.data[
            "launchPipelineReexecution"
        ]
        run_id = retry.data["launchPipelineReexecution"]["run"]["runId"]
        run = graphql_context.instance.get_run_by_id(run_id)
        assert run and run.status == DagsterRunStatus.SUCCESS
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            "pipelineRunLogs"
        ]["messages"]
        assert step_did_succeed(logs, "always_succeed")
        assert step_did_succeed(logs, "conditionally_fail")
        assert step_did_succeed(logs, "after_failure")

    def test_retry_asset_job_with_reexecution_params(
        self, graphql_context: WorkspaceRequestContext
    ):
        selector = infer_job_selector(
            graphql_context, "two_assets_job", asset_selection=[{"path": ["asset_one"]}]
        )
        result = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                }
            },
        )

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            "pipelineRunLogs"
        ]["messages"]

        assert step_did_succeed(logs, "asset_one")
        assert step_did_not_run(logs, "asset_two")

        retry_one = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={"reexecutionParams": {"parentRunId": run_id, "strategy": "ALL_STEPS"}},
        )
        run_id = retry_one.data["launchPipelineReexecution"]["run"]["runId"]
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            "pipelineRunLogs"
        ]["messages"]
        assert step_did_succeed(logs, "asset_one")
        assert step_did_not_run(logs, "asset_two")

    def test_retry_hard_failure_with_reexecution_params_run_config_changed(
        self, graphql_context: WorkspaceRequestContext
    ):
        """Test that reexecution fails if the run config changes."""
        selector = infer_job_selector(graphql_context, "chained_failure_job")

        # trigger failure in the conditionally_fail op
        output_file = os.path.join(
            get_system_temp_directory(), "chained_failure_job_conditionally_fail"
        )
        try:
            with open(output_file, "w", encoding="utf8"):
                result = execute_dagster_graphql_and_finish_runs(
                    graphql_context,
                    LAUNCH_PIPELINE_EXECUTION_MUTATION,
                    variables={
                        "executionParams": {
                            "selector": selector,
                        }
                    },
                )
        finally:
            os.remove(output_file)

        parent_run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        parent_run = graphql_context.instance.get_run_by_id(parent_run_id)
        assert parent_run
        assert parent_run.status == DagsterRunStatus.FAILURE

        # override run config to make it fail
        graphql_context.instance.delete_run(parent_run_id)
        graphql_context.instance.add_run(parent_run._replace(run_config={"bad": "config"}))

        retry = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "reexecutionParams": {"parentRunId": parent_run_id, "strategy": "FROM_FAILURE"}
            },
        )
        assert "DagsterInvalidConfigError" in str(
            retry.data["launchPipelineReexecution"]["message"]
        )

    def test_retry_failure_with_reexecution_params(self, graphql_context: WorkspaceRequestContext):
        """Test with providng reexecutionParams rather than executionParams."""
        selector = infer_job_selector(graphql_context, "chained_failure_job")

        # trigger failure in the conditionally_fail op
        output_file = os.path.join(
            get_system_temp_directory(), "chained_failure_job_conditionally_fail"
        )
        try:
            with open(output_file, "w", encoding="utf8"):
                result = execute_dagster_graphql_and_finish_runs(
                    graphql_context,
                    LAUNCH_PIPELINE_EXECUTION_MUTATION,
                    variables={
                        "executionParams": {
                            "selector": selector,
                        }
                    },
                )
        finally:
            os.remove(output_file)

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        run = graphql_context.instance.get_run_by_id(run_id)
        assert run and run.status == DagsterRunStatus.FAILURE

        retry = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={"reexecutionParams": {"parentRunId": run_id, "strategy": "FROM_FAILURE"}},
        )

        run_id = retry.data["launchPipelineReexecution"]["run"]["runId"]
        run = graphql_context.instance.get_run_by_id(run_id)
        assert run and run.status == DagsterRunStatus.SUCCESS
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            "pipelineRunLogs"
        ]["messages"]
        assert step_did_not_run(logs, "always_succeed")
        assert step_did_succeed(logs, "conditionally_fail")
        assert step_did_succeed(logs, "after_failure")


def test_graphene_reexecution_strategy():
    """Check that graphene enum has corresponding values in the ReexecutionStrategy enum."""
    for strategy in GrapheneReexecutionStrategy.__enum__:
        assert ReexecutionStrategy[strategy.value]


def _do_retry_intermediates_test(graphql_context):
    selector = infer_job_selector(graphql_context, "eventually_successful")
    result = execute_dagster_graphql_and_finish_runs(
        graphql_context,
        LAUNCH_PIPELINE_EXECUTION_MUTATION,
        variables={
            "executionParams": {
                "selector": selector,
            }
        },
    )
    assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
    run_id = result.data["launchPipelineExecution"]["run"]["runId"]
    logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
        "pipelineRunLogs"
    ]["messages"]

    assert step_did_succeed(logs, "spawn")
    assert step_did_fail(logs, "fail")
    assert step_did_not_run(logs, "fail_2")
    assert step_did_not_run(logs, "fail_3")
    assert step_did_not_run(logs, "reset")

    retry_one = execute_dagster_graphql_and_finish_runs(
        graphql_context,
        LAUNCH_PIPELINE_REEXECUTION_MUTATION,
        variables={
            "executionParams": {
                "selector": selector,
                "executionMetadata": {
                    "rootRunId": run_id,
                    "parentRunId": run_id,
                    "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                },
            }
        },
    )

    retry_run_id = retry_one.data["launchPipelineReexecution"]["run"]["runId"]
    return retry_run_id


class TestRetryExecutionAsyncOnlyBehavior(ExecutingGraphQLContextTestMatrix):
    def test_retry_requires_intermediates_async_only(
        self, graphql_context: WorkspaceRequestContext
    ):
        # run_id = make_new_run_id()
        # reexecution_run_id = make_new_run_id()

        reexecution_run_id = _do_retry_intermediates_test(graphql_context)
        reexecution_run = graphql_context.instance.get_run_by_id(reexecution_run_id)
        assert reexecution_run
        assert reexecution_run.is_failure_or_canceled

    def test_retry_early_terminate(self, graphql_context: WorkspaceRequestContext):
        instance = graphql_context.instance
        selector = infer_job_selector(graphql_context, "retry_multi_input_early_terminate_job")
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": {
                        "ops": {
                            "get_input_one": {"config": {"wait_to_terminate": True}},
                            "get_input_two": {"config": {"wait_to_terminate": True}},
                        },
                    },
                }
            },
        )

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        # Wait until the first step succeeded
        while instance.get_run_stats(run_id).steps_succeeded < 1:
            sleep(0.1)
        # Terminate the current pipeline run at the second step
        graphql_context.instance.run_launcher.terminate(run_id)

        records = instance.all_logs(run_id)

        # The first step should succeed, the second should fail or not start,
        # and the following steps should not appear in records
        assert step_did_succeed_in_records(records, "return_one")
        assert not step_did_fail_in_records(records, "return_one")
        assert any(
            [
                step_did_fail_in_records(records, "get_input_one"),
                step_did_not_run_in_records(records, "get_input_one"),
            ]
        )
        assert step_did_not_run_in_records(records, "get_input_two")
        assert step_did_not_run_in_records(records, "sum_inputs")

        # Wait for the original run to finish
        poll_for_finished_run(instance, run_id, timeout=30)
        run = instance.get_run_by_id(run_id)
        assert run and run.status == DagsterRunStatus.CANCELED

        # Start retry
        retry_result = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": {
                        "ops": {
                            "get_input_one": {"config": {"wait_to_terminate": False}},
                            "get_input_two": {"config": {"wait_to_terminate": False}},
                        },
                    },
                    "executionMetadata": {
                        "rootRunId": run_id,
                        "parentRunId": run_id,
                        "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                    },
                }
            },
        )

        new_run_id = retry_result.data["launchPipelineReexecution"]["run"]["runId"]
        retry_records = instance.all_logs(new_run_id)
        # The first step should not run and the other three steps should succeed in retry
        assert step_did_not_run_in_records(retry_records, "return_one")
        assert step_did_succeed_in_records(retry_records, "get_input_one")
        assert step_did_succeed_in_records(retry_records, "get_input_two")
        assert step_did_succeed_in_records(retry_records, "sum_inputs")



import gc
import re
from contextlib import ExitStack
from unittest import mock

import pytest
from dagster import job, op, repository
from dagster._core.remote_representation.code_location import GrpcServerCodeLocation
from dagster._core.test_utils import instance_for_test
from dagster_graphql.test.utils import define_out_of_process_workspace, main_repo_location_name


def get_repo():
    """This is a repo that changes name very time it's loaded."""

    @op
    def op_A():
        pass

    @job
    def pipe():
        op_A()

    import random
    import string

    @repository(name="".join(random.choice(string.ascii_lowercase) for i in range(10)))
    def my_repo():
        return [pipe]

    return my_repo


def test_can_reload_on_remote_repository_error():
    with instance_for_test() as instance:
        with ExitStack() as exit_stack:
            with mock.patch(
                # note it where the function is *used* that needs to mocked, not
                # where it is defined.
                # see https://docs.python.org/3/library/unittest.mock.html#where-to-patch
                "dagster._api.snapshot_repository.sync_get_external_repositories_data_grpc"
            ) as remote_repository_mock:
                remote_repository_mock.side_effect = Exception("get_remote_repo_failure")

                with pytest.warns(UserWarning, match=re.escape("get_remote_repo_failure")):
                    workspace = exit_stack.enter_context(
                        define_out_of_process_workspace(__file__, "get_repo", instance)
                    )

                assert not workspace.has_code_location(main_repo_location_name())
                assert workspace.has_code_location_error(main_repo_location_name())

            workspace.reload_code_location(main_repo_location_name())
            assert workspace.has_code_location(main_repo_location_name())


def test_reload_on_process_context():
    with instance_for_test() as instance:
        with define_out_of_process_workspace(__file__, "get_repo", instance) as process_context:
            request_context = process_context.create_request_context()

            # Save the repository name
            code_location = request_context.code_locations[0]
            repo = next(iter(code_location.get_repositories().values()))
            repo_name = repo.name

            # Reload the location and save the new repository name
            process_context.reload_code_location(code_location.name)

            new_request_context = process_context.create_request_context()

            code_location = new_request_context.code_locations[0]
            repo = next(iter(code_location.get_repositories().values()))
            new_repo_name = repo.name

            # Check that the repository has changed
            assert repo_name != new_repo_name


def test_reload_on_request_context():
    with instance_for_test() as instance:
        with define_out_of_process_workspace(__file__, "get_repo", instance) as process_context:
            assert process_context.code_locations_count == 1

            # Create a request context from the process context
            request_context = process_context.create_request_context()

            # Save the repository name
            code_location = request_context.code_locations[0]
            repo = next(iter(code_location.get_repositories().values()))
            repo_name = repo.name

            # Reload the location and save the new repository name
            process_context.reload_code_location(code_location.name)

            new_request_context = process_context.create_request_context()
            code_location = new_request_context.code_locations[0]
            repo = next(iter(code_location.get_repositories().values()))
            new_repo_name = repo.name

            # Check that the repository has changed
            assert repo_name != new_repo_name

            # Check that the repository name is still the same on the old request context,
            # confirming that the old repository location is still running
            code_location = request_context.code_locations[0]
            repo = next(iter(code_location.get_repositories().values()))
            assert repo_name == repo.name


def test_reload_on_request_context_2():
    # This is the similar to the test `test_reload_on_request_context`,
    # but calls reload from the request_context instead of on the process_context

    with instance_for_test() as instance:
        with define_out_of_process_workspace(__file__, "get_repo", instance) as process_context:
            assert process_context.code_locations_count == 1

            request_context = process_context.create_request_context()

            # Save the repository name
            code_location = request_context.code_locations[0]
            repo = next(iter(code_location.get_repositories().values()))
            repo_name = repo.name

            # Reload the location from the request context
            new_request_context = request_context.reload_code_location(code_location.name)

            code_location = new_request_context.code_locations[0]
            repo = next(iter(code_location.get_repositories().values()))
            new_repo_name = repo.name

            # Check that the repository has changed
            assert repo_name != new_repo_name

            # Check that the repository name is still the same on the old request context,
            # confirming that the old repository location is still running
            code_location = request_context.code_locations[0]
            repo = next(iter(code_location.get_repositories().values()))
            assert repo_name == repo.name


def test_handle_cleanup_by_workspace_context_exit():
    with instance_for_test() as instance:
        with mock.patch.object(GrpcServerCodeLocation, "cleanup") as mock_method:
            with define_out_of_process_workspace(__file__, "get_repo", instance):
                pass

    assert mock_method.called


def test_handle_cleaup_by_gc_without_request_context():
    called = {"yup": False}

    def call_me():
        called["yup"] = True

    with instance_for_test() as instance:
        with define_out_of_process_workspace(__file__, "get_repo", instance) as process_context:
            assert process_context.code_locations_count == 1

            request_context = process_context.create_request_context()
            request_context.code_locations[0].cleanup = call_me

            # Reload the location from the request context
            assert not called["yup"]
            process_context.reload_code_location("test_location")
            assert not called["yup"]

            request_context = None

            # There are no more references to the location, so it should be GC'd
            gc.collect()
            assert called["yup"]



from dagster._core.launcher.sync_in_memory_run_launcher import SyncInMemoryRunLauncher
from dagster._core.test_utils import create_run_for_test, instance_for_test

from dagster_graphql_tests.graphql.repo import (
    get_main_workspace,
    main_repo_location_name,
    main_repo_name,
)


def test_sync_run_launcher_from_configurable_class():
    with instance_for_test(synchronous_run_launcher=True) as instance_no_hijack:
        assert isinstance(instance_no_hijack.run_launcher, SyncInMemoryRunLauncher)


def test_sync_run_launcher_run():
    with instance_for_test(synchronous_run_launcher=True) as instance:
        with get_main_workspace(instance) as workspace:
            location = workspace.get_code_location(main_repo_location_name())
            repo = location.get_repository(main_repo_name())
            job = repo.get_full_job("noop_job")

            run = create_run_for_test(
                instance=instance,
                job_name=job.name,
                remote_job_origin=job.get_remote_origin(),
                job_code_origin=job.get_python_origin(),
            )

            run = instance.launch_run(run_id=run.run_id, workspace=workspace)

            completed_run = instance.get_run_by_id(run.run_id)
            assert completed_run.is_success  # pyright: ignore[reportOptionalMemberAccess]



import sys
import time
from typing import Any
from unittest import mock

import pytest
from dagster import file_relative_path
from dagster._core.remote_origin import ManagedGrpcPythonEnvCodeLocationOrigin
from dagster._core.remote_representation.feature_flags import CodeLocationFeatureFlags
from dagster._core.test_utils import environ
from dagster._core.types.loadable_target_origin import LoadableTargetOrigin
from dagster._core.workspace.load import location_origins_from_yaml_paths
from dagster.version import __version__ as dagster_version
from dagster_graphql.test.utils import execute_dagster_graphql, main_repo_location_name
from dagster_graphql.version import __version__ as dagster_graphql_version
from dagster_shared.version import __version__ as dagster_shared_version

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    GraphQLContextVariant,
    make_graphql_context_test_suite,
)

LOCATION_ENTRY_FRAGMENT = """
fragment locationEntryFragment on WorkspaceLocationEntry {
  __typename
  id
  name
  definitionsSource
  locationOrLoadError {
    __typename
    ... on RepositoryLocation {
      id
      name
      repositories {
        name
      }
      isReloadSupported
      dagsterLibraryVersions {
        name
        version
      }
    }
    ... on PythonError {
      message
      stack
      errorChain {
        error {
          message
          stack
        }
      }
    }
  }
  loadStatus
  displayMetadata {
    key
    value
  }
  updatedTimestamp
  featureFlags {
    name
    enabled
  }
}
"""

WORKSPACE_QUERY = (
    """
query {
  workspaceOrError {
    __typename
    ... on Workspace {
      locationEntries {
        ...locationEntryFragment
      }
    }
    ... on PythonError {
      message
      stack
      errorChain {
        error {
          message
          stack
        }
      }
    }
  }
}
"""
    + LOCATION_ENTRY_FRAGMENT
)

LOCATION_STATUS_QUERY = """
query {
   locationStatusesOrError {
      __typename
      ... on WorkspaceLocationStatusEntries {
        entries {
          __typename
          id
          loadStatus
          updateTimestamp
        }
      }
      ... on PythonError {
          message
          stack
      }
    }
}
"""

LOCATION_ENTRY_QUERY = (
    """
query ($name: String!) {
  workspaceLocationEntryOrError(name: $name) {
    __typename
    ...locationEntryFragment
    ... on PythonError {
        message
        stack
    }
  }
}
"""
    + LOCATION_ENTRY_FRAGMENT
)


@pytest.fixture(scope="function")
def enable_masking_user_code_errors() -> Any:
    with environ({"DAGSTER_REDACT_USER_CODE_ERRORS": "1"}):
        yield


BaseTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[GraphQLContextVariant.non_launchable_sqlite_instance_multi_location()]
)


class TestLoadWorkspace(BaseTestSuite):
    def test_load_workspace(self, graphql_context):
        # Add an error origin
        original_origins = location_origins_from_yaml_paths(
            [file_relative_path(__file__, "multi_location.yaml")]
        )
        with mock.patch(
            "dagster._core.workspace.load_target.location_origins_from_yaml_paths",
        ) as origins_mock:
            original_origins.append(  # pyright: ignore[reportAttributeAccessIssue]
                ManagedGrpcPythonEnvCodeLocationOrigin(
                    location_name="error_location",
                    loadable_target_origin=LoadableTargetOrigin(
                        python_file="made_up_file.py", executable_path=sys.executable
                    ),
                )
            )

            origins_mock.return_value = original_origins

            reload_time = time.time()

            new_context = graphql_context.reload_workspace()

            result = execute_dagster_graphql(new_context, WORKSPACE_QUERY)

            assert result
            assert result.data
            assert result.data["workspaceOrError"]
            assert result.data["workspaceOrError"]["__typename"] == "Workspace", str(result.data)

            nodes = result.data["workspaceOrError"]["locationEntries"]

            assert len(nodes) == 3

            assert all([node["__typename"] == "WorkspaceLocationEntry" for node in nodes]), str(
                nodes
            )
            assert all([node["definitionsSource"] == "CODE_SERVER" for node in nodes]), str(nodes)

            success_nodes = [
                node["locationOrLoadError"]
                for node in nodes
                if node["locationOrLoadError"]["__typename"] == "RepositoryLocation"
            ]
            assert len(success_nodes) == 2
            assert success_nodes[0]["dagsterLibraryVersions"] == [
                {"name": "dagster", "version": dagster_version},
                {"name": "dagster-graphql", "version": dagster_graphql_version},
                {"name": "dagster-shared", "version": dagster_shared_version},
            ]

            failures = [
                node for node in nodes if node["locationOrLoadError"]["__typename"] == "PythonError"
            ]
            assert len(failures) == 1
            failure_node = failures[0]

            assert failure_node["name"] == "error_location"
            assert failure_node["loadStatus"] == "LOADED"

            assert "No such file or directory" in str(failure_node["locationOrLoadError"]), (
                failure_node
            )

            for node in nodes:
                assert node["loadStatus"] == "LOADED"
                update_time = node["updatedTimestamp"]
                assert update_time >= reload_time and update_time <= time.time()

                metadatas = node["displayMetadata"]
                metadata_dict = {metadata["key"]: metadata["value"] for metadata in metadatas}

                feature_flags = node["featureFlags"]
                assert len(feature_flags) == 1
                assert (
                    feature_flags[0]["name"]
                    == CodeLocationFeatureFlags.SHOW_SINGLE_RUN_BACKFILL_TOGGLE.value
                )

                assert (
                    "python_file" in metadata_dict
                    or "module_name" in metadata_dict
                    or "package_name" in metadata_dict
                )

    def test_load_location_statuses(self, graphql_context):
        original_origins = location_origins_from_yaml_paths(
            [file_relative_path(__file__, "multi_location.yaml")]
        )
        with mock.patch(
            "dagster._core.workspace.load_target.location_origins_from_yaml_paths",
        ) as origins_mock:
            # Add an error origin
            original_origins.append(  # pyright: ignore[reportAttributeAccessIssue]
                ManagedGrpcPythonEnvCodeLocationOrigin(
                    location_name="error_location",
                    loadable_target_origin=LoadableTargetOrigin(
                        python_file="made_up_file.py", executable_path=sys.executable
                    ),
                )
            )

            origins_mock.return_value = original_origins

            reload_timestamp = time.time()

            new_context = graphql_context.reload_workspace()

            result = execute_dagster_graphql(new_context, LOCATION_STATUS_QUERY)

            assert result
            assert result.data
            assert result.data["locationStatusesOrError"]
            assert (
                result.data["locationStatusesOrError"]["__typename"]
                == "WorkspaceLocationStatusEntries"
            ), str(result.data)

            nodes = result.data["locationStatusesOrError"]["entries"]

            assert len(nodes) == 3

            assert all([node["__typename"] == "WorkspaceLocationStatusEntry" for node in nodes]), (
                str(nodes)
            )

            for node in nodes:
                assert node["loadStatus"] == "LOADED"
                assert float(node["updateTimestamp"]) > reload_timestamp

    def test_load_workspace_masked(self, graphql_context, enable_masking_user_code_errors):
        # Add an error origin
        original_origins = location_origins_from_yaml_paths(
            [file_relative_path(__file__, "multi_location.yaml")]
        )
        with mock.patch(
            "dagster._core.workspace.load_target.location_origins_from_yaml_paths",
        ) as origins_mock:
            original_origins.append(  # pyright: ignore[reportAttributeAccessIssue]
                ManagedGrpcPythonEnvCodeLocationOrigin(
                    location_name="error_location",
                    loadable_target_origin=LoadableTargetOrigin(
                        python_file="made_up_file.py", executable_path=sys.executable
                    ),
                )
            )

            origins_mock.return_value = original_origins

            new_context = graphql_context.reload_workspace()

            result = execute_dagster_graphql(new_context, WORKSPACE_QUERY)

            assert result
            assert result.data
            assert result.data["workspaceOrError"]
            assert result.data["workspaceOrError"]["__typename"] == "Workspace", str(result.data)

            nodes = result.data["workspaceOrError"]["locationEntries"]

            assert len(nodes) == 3

            assert all([node["__typename"] == "WorkspaceLocationEntry" for node in nodes]), str(
                nodes
            )

            success_nodes = [
                node["locationOrLoadError"]
                for node in nodes
                if node["locationOrLoadError"]["__typename"] == "RepositoryLocation"
            ]
            assert len(success_nodes) == 2
            assert success_nodes[0]["dagsterLibraryVersions"] == [
                {"name": "dagster", "version": dagster_version},
                {"name": "dagster-graphql", "version": dagster_graphql_version},
                {"name": "dagster-shared", "version": dagster_shared_version},
            ]

            failures = [
                node for node in nodes if node["locationOrLoadError"]["__typename"] == "PythonError"
            ]
            assert len(failures) == 1
            failure_node = failures[0]

            assert failure_node["name"] == "error_location"
            assert failure_node["loadStatus"] == "LOADED"
            assert "No such file or directory" not in str(failure_node["locationOrLoadError"]), (
                failure_node["locationOrLoadError"]["message"]
            )
            assert (
                "Search in logs for this error ID for more details"
                in failure_node["locationOrLoadError"]["message"]
            )

    def test_workspace_entry_by_name(self, graphql_context) -> None:
        result = execute_dagster_graphql(
            graphql_context, LOCATION_ENTRY_QUERY, {"name": main_repo_location_name()}
        )
        assert result
        assert result.data["workspaceLocationEntryOrError"]
        assert (
            result.data["workspaceLocationEntryOrError"]["__typename"] == "WorkspaceLocationEntry"
        )
        assert (
            result.data["workspaceLocationEntryOrError"]["locationOrLoadError"]["__typename"]
            == "RepositoryLocation"
        )

        result = execute_dagster_graphql(
            graphql_context, LOCATION_ENTRY_QUERY, {"name": "does_not_exist"}
        )
        assert result
        assert result.data["workspaceLocationEntryOrError"] is None

        with mock.patch(
            "dagster_graphql.schema.roots.query.fetch_location_entry",
        ) as mock_fetch:
            mock_fetch.side_effect = Exception("boom")

            result = execute_dagster_graphql(
                graphql_context,
                LOCATION_ENTRY_QUERY,
                {"name": main_repo_location_name()},
            )
            assert result
            assert result.data["workspaceLocationEntryOrError"]
            assert result.data["workspaceLocationEntryOrError"]["__typename"] == "PythonError"
            assert "boom" in result.data["workspaceLocationEntryOrError"]["message"]



import sys
from typing import TYPE_CHECKING

import dagster._check as check
import graphene
import yaml
from dagster._core.instance import DagsterInstance
from dagster._core.instance.config import PoolConfig
from dagster._core.launcher.base import RunLauncher
from dagster._core.storage.event_log.sql_event_log import SqlEventLogStorage
from dagster._daemon.asset_daemon import get_auto_materialize_paused
from dagster._daemon.types import DaemonStatus
from dagster._utils.concurrency import (
    ClaimedSlotInfo,
    PendingStepInfo,
    get_max_concurrency_limit_value,
)

from dagster_graphql.schema.errors import GraphenePythonError
from dagster_graphql.schema.util import ResolveInfo, non_null_list

if TYPE_CHECKING:
    from dagster._core.run_coordinator.queued_run_coordinator import RunQueueConfig


class GrapheneRunLauncher(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)

    class Meta:
        name = "RunLauncher"

    def __init__(self, run_launcher):
        super().__init__()
        self._run_launcher = check.inst_param(run_launcher, "run_launcher", RunLauncher)

    def resolve_name(self, _graphene_info: ResolveInfo):
        return self._run_launcher.__class__.__name__


class GrapheneDaemonStatus(graphene.ObjectType):
    daemonType = graphene.NonNull(graphene.String)
    id = graphene.NonNull(graphene.ID)
    required = graphene.NonNull(graphene.Boolean)
    healthy = graphene.Boolean()
    lastHeartbeatTime = graphene.Float()
    lastHeartbeatErrors = non_null_list(GraphenePythonError)

    class Meta:
        name = "DaemonStatus"

    def __init__(self, daemon_status):
        check.inst_param(daemon_status, "daemon_status", DaemonStatus)

        super().__init__(
            daemonType=daemon_status.daemon_type,
            required=daemon_status.required,
            healthy=daemon_status.healthy,
            lastHeartbeatTime=(
                daemon_status.last_heartbeat.timestamp if daemon_status.last_heartbeat else None
            ),
            lastHeartbeatErrors=(
                [GraphenePythonError(error) for error in daemon_status.last_heartbeat.errors]
                if daemon_status.last_heartbeat and daemon_status.last_heartbeat.errors
                else []
            ),
        )

    def resolve_id(self, _graphene_info: ResolveInfo):
        return self.daemonType


class GrapheneDaemonHealth(graphene.ObjectType):
    id = graphene.NonNull(graphene.String)
    daemonStatus = graphene.Field(
        graphene.NonNull(GrapheneDaemonStatus),
        daemon_type=graphene.Argument(graphene.String),
    )
    allDaemonStatuses = non_null_list(GrapheneDaemonStatus)

    class Meta:
        name = "DaemonHealth"

    def __init__(self, instance):
        super().__init__()
        self._instance = check.inst_param(instance, "instance", DagsterInstance)

    def resolve_id(self, _graphene_info: ResolveInfo):
        return "daemonHealth"

    def resolve_daemonStatus(self, _graphene_info: ResolveInfo, daemon_type):
        check.str_param(daemon_type, "daemon_type")
        return GrapheneDaemonStatus(
            self._instance.get_daemon_statuses(daemon_types=[daemon_type])[daemon_type]
        )

    def resolve_allDaemonStatuses(self, _graphene_info: ResolveInfo):
        return [
            GrapheneDaemonStatus(daemon_status)
            for daemon_status in self._instance.get_daemon_statuses().values()
        ]


class GrapheneClaimedConcurrencySlot(graphene.ObjectType):
    runId = graphene.NonNull(graphene.String)
    stepKey = graphene.NonNull(graphene.String)

    class Meta:
        name = "ClaimedConcurrencySlot"

    def __init__(self, claimed_slot_info: ClaimedSlotInfo):
        super().__init__(runId=claimed_slot_info.run_id, stepKey=claimed_slot_info.step_key)


class GraphenePendingConcurrencyStep(graphene.ObjectType):
    runId = graphene.NonNull(graphene.String)
    stepKey = graphene.NonNull(graphene.String)
    enqueuedTimestamp = graphene.NonNull(graphene.Float)
    assignedTimestamp = graphene.Float()
    priority = graphene.Int()

    class Meta:
        name = "PendingConcurrencyStep"

    def __init__(self, pending_step_info: PendingStepInfo):
        super().__init__(
            runId=pending_step_info.run_id,
            stepKey=pending_step_info.step_key,
            enqueuedTimestamp=pending_step_info.enqueued_timestamp.timestamp(),
            assignedTimestamp=pending_step_info.assigned_timestamp.timestamp()
            if pending_step_info.assigned_timestamp
            else None,
            priority=pending_step_info.priority,
        )


class GrapheneConcurrencyKeyInfo(graphene.ObjectType):
    concurrencyKey = graphene.NonNull(graphene.String)
    slotCount = graphene.NonNull(graphene.Int)
    claimedSlots = non_null_list(GrapheneClaimedConcurrencySlot)
    pendingSteps = non_null_list(GraphenePendingConcurrencyStep)
    activeSlotCount = graphene.NonNull(graphene.Int)
    activeRunIds = non_null_list(graphene.String)
    pendingStepCount = graphene.NonNull(graphene.Int)
    pendingStepRunIds = non_null_list(graphene.String)
    assignedStepCount = graphene.NonNull(graphene.Int)
    assignedStepRunIds = non_null_list(graphene.String)
    limit = graphene.Int()
    usingDefaultLimit = graphene.Boolean()

    class Meta:
        name = "ConcurrencyKeyInfo"

    def __init__(self, concurrency_key: str):
        self._concurrency_key = concurrency_key
        self._concurrency_key_info = None
        super().__init__(concurrencyKey=concurrency_key)

    def _get_concurrency_key_info(self, graphene_info: ResolveInfo):
        if not self._concurrency_key_info:
            self._concurrency_key_info = (
                graphene_info.context.instance.event_log_storage.get_concurrency_info(
                    self._concurrency_key
                )
            )
        return self._concurrency_key_info

    def resolve_claimedSlots(self, graphene_info: ResolveInfo):
        return [
            GrapheneClaimedConcurrencySlot(slot)
            for slot in self._get_concurrency_key_info(graphene_info).claimed_slots
        ]

    def resolve_pendingSteps(self, graphene_info: ResolveInfo):
        return [
            GraphenePendingConcurrencyStep(step)
            for step in self._get_concurrency_key_info(graphene_info).pending_steps
        ]

    def resolve_slotCount(self, graphene_info: ResolveInfo):
        return self._get_concurrency_key_info(graphene_info).slot_count

    def resolve_activeSlotCount(self, graphene_info: ResolveInfo):
        return self._get_concurrency_key_info(graphene_info).active_slot_count

    def resolve_activeRunIds(self, graphene_info: ResolveInfo):
        return list(self._get_concurrency_key_info(graphene_info).active_run_ids)

    def resolve_pendingStepCount(self, graphene_info: ResolveInfo):
        return self._get_concurrency_key_info(graphene_info).pending_step_count

    def resolve_pendingStepRunIds(self, graphene_info: ResolveInfo):
        return list(self._get_concurrency_key_info(graphene_info).pending_run_ids)

    def resolve_assignedStepCount(self, graphene_info: ResolveInfo):
        return self._get_concurrency_key_info(graphene_info).assigned_step_count

    def resolve_assignedStepRunIds(self, graphene_info: ResolveInfo):
        return list(self._get_concurrency_key_info(graphene_info).assigned_run_ids)

    def resolve_limit(self, graphene_info: ResolveInfo):
        return self._get_concurrency_key_info(graphene_info).limit

    def resolve_usingDefaultLimit(self, graphene_info: ResolveInfo):
        return self._get_concurrency_key_info(graphene_info).using_default_limit


class GrapheneRunQueueConfig(graphene.ObjectType):
    maxConcurrentRuns = graphene.NonNull(graphene.Int)
    tagConcurrencyLimitsYaml = graphene.String()
    isOpConcurrencyAware = graphene.Boolean()

    class Meta:
        name = "RunQueueConfig"

    def __init__(self, run_queue_config: "RunQueueConfig"):
        super().__init__()
        self._run_queue_config = run_queue_config

    def resolve_maxConcurrentRuns(self, _graphene_info: ResolveInfo):
        return self._run_queue_config.max_concurrent_runs

    def resolve_tagConcurrencyLimitsYaml(self, _graphene_info: ResolveInfo):
        if not self._run_queue_config.tag_concurrency_limits:
            return None

        return yaml.dump(
            self._run_queue_config.tag_concurrency_limits,
            default_flow_style=False,
            allow_unicode=True,
            sort_keys=True,
        )

    def resolve_isOpConcurrencyAware(self, _graphene_info: ResolveInfo):
        return self._run_queue_config.should_block_op_concurrency_limited_runs


class GraphenePoolConfig(graphene.ObjectType):
    poolGranularity = graphene.String()
    defaultPoolLimit = graphene.Int()
    opGranularityRunBuffer = graphene.Int()

    class Meta:
        name = "PoolConfig"

    def __init__(self, pool_config: PoolConfig):
        super().__init__()
        self._pool_config = check.inst_param(pool_config, "pool_config", PoolConfig)

    def resolve_poolGranularity(self, _graphene_info: ResolveInfo):
        return (
            self._pool_config.pool_granularity.value if self._pool_config.pool_granularity else None
        )

    def resolve_defaultPoolLimit(self, _graphene_info: ResolveInfo):
        return self._pool_config.default_pool_limit

    def resolve_opGranularityRunBuffer(self, _graphene_info: ResolveInfo):
        return self._pool_config.op_granularity_run_buffer


class GrapheneInstance(graphene.ObjectType):
    id = graphene.NonNull(graphene.String)
    info = graphene.Field(graphene.String)
    runLauncher = graphene.Field(GrapheneRunLauncher)
    runQueuingSupported = graphene.NonNull(graphene.Boolean)
    runQueueConfig = graphene.Field(GrapheneRunQueueConfig)
    executablePath = graphene.NonNull(graphene.String)
    daemonHealth = graphene.NonNull(GrapheneDaemonHealth)
    hasInfo = graphene.NonNull(graphene.Boolean)
    autoMaterializePaused = graphene.NonNull(graphene.Boolean)
    supportsConcurrencyLimits = graphene.NonNull(graphene.Boolean)
    minConcurrencyLimitValue = graphene.NonNull(graphene.Int)
    maxConcurrencyLimitValue = graphene.NonNull(graphene.Int)
    concurrencyLimits = non_null_list(GrapheneConcurrencyKeyInfo)
    concurrencyLimit = graphene.Field(
        graphene.NonNull(GrapheneConcurrencyKeyInfo),
        concurrencyKey=graphene.Argument(graphene.String),
    )
    useAutoMaterializeSensors = graphene.Field(
        graphene.NonNull(graphene.Boolean),
        description="Whether or not the deployment is using automation policy sensors to materialize assets",
    )
    poolConfig = graphene.Field(GraphenePoolConfig)
    freshnessEvaluationEnabled = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "Instance"

    def __init__(self, instance):
        super().__init__()
        self._instance = check.inst_param(instance, "instance", DagsterInstance)

    def resolve_id(self, _graphene_info: ResolveInfo):
        return "Instance"

    def resolve_hasInfo(self, graphene_info: ResolveInfo) -> bool:
        return graphene_info.context.show_instance_config

    def resolve_useAutoMaterializeSensors(self, _graphene_info: ResolveInfo):
        return self._instance.auto_materialize_use_sensors

    def resolve_info(self, graphene_info: ResolveInfo):
        return self._instance.info_str() if graphene_info.context.show_instance_config else None

    def resolve_runLauncher(self, _graphene_info: ResolveInfo):
        return (
            GrapheneRunLauncher(self._instance.run_launcher)
            if self._instance.run_launcher
            else None
        )

    def resolve_runQueuingSupported(self, _graphene_info: ResolveInfo):
        from dagster._core.run_coordinator import QueuedRunCoordinator

        return isinstance(self._instance.run_coordinator, QueuedRunCoordinator)

    def resolve_runQueueConfig(self, _graphene_info: ResolveInfo):
        run_queue_config = self._instance.get_concurrency_config().run_queue_config
        if run_queue_config:
            return GrapheneRunQueueConfig(run_queue_config)
        else:
            return None

    def resolve_executablePath(self, _graphene_info: ResolveInfo):
        return sys.executable

    def resolve_daemonHealth(self, _graphene_info: ResolveInfo):
        return GrapheneDaemonHealth(instance=self._instance)

    def resolve_autoMaterializePaused(self, _graphene_info: ResolveInfo):
        return get_auto_materialize_paused(self._instance)

    def resolve_supportsConcurrencyLimits(self, _graphene_info: ResolveInfo):
        return self._instance.event_log_storage.supports_global_concurrency_limits

    def resolve_concurrencyLimits(self, _graphene_info: ResolveInfo):
        res = []
        for key in self._instance.event_log_storage.get_concurrency_keys():
            res.append(GrapheneConcurrencyKeyInfo(key))
        return res

    def resolve_concurrencyLimit(self, _graphene_info: ResolveInfo, concurrencyKey):
        return GrapheneConcurrencyKeyInfo(concurrencyKey)

    def resolve_minConcurrencyLimitValue(self, _graphene_info: ResolveInfo):
        if isinstance(
            self._instance.event_log_storage, SqlEventLogStorage
        ) and not self._instance.event_log_storage.has_table("concurrency_limits"):
            return 1
        return 0

    def resolve_maxConcurrencyLimitValue(self, _graphene_info: ResolveInfo):
        return get_max_concurrency_limit_value()

    def resolve_poolConfig(self, _graphene_info: ResolveInfo):
        concurrency_config = self._instance.get_concurrency_config()
        return GraphenePoolConfig(concurrency_config.pool_config)

    def resolve_freshnessEvaluationEnabled(self, graphene_info: ResolveInfo):
        return (
            # freshness daemon turned on
            graphene_info.context.instance.freshness_enabled
            # override
            or graphene_info.context.instance.internal_asset_freshness_enabled()
        )



from unittest import mock

from dagster import AssetKey, AssetOut, Definitions, asset, multi_asset
from dagster._core.definitions.assets.graph.asset_graph_subset import AssetGraphSubset
from dagster._core.definitions.partitions.definition import (
    DailyPartitionsDefinition,
    DynamicPartitionsDefinition,
    HourlyPartitionsDefinition,
    StaticPartitionsDefinition,
)
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
)
from dagster._core.execution.asset_backfill import AssetBackfillData
from dagster._core.instance import DagsterInstance
from dagster._core.test_utils import ensure_dagster_tests_import, instance_for_test
from dagster._daemon import get_default_daemon_logger
from dagster._daemon.backfill import execute_backfill_iteration
from dagster_graphql.client.query import LAUNCH_PARTITION_BACKFILL_MUTATION
from dagster_graphql.test.utils import (
    GqlResult,
    define_out_of_process_context,
    execute_dagster_graphql,
    main_repo_location_name,
)

ensure_dagster_tests_import()
from dagster_tests.declarative_automation_tests.legacy_tests.scenarios.asset_graphs import (
    root_assets_different_partitions_same_downstream,
)

GET_PARTITION_BACKFILLS_QUERY = """
  query InstanceBackfillsQuery($cursor: String, $limit: Int) {
    partitionBackfillsOrError(cursor: $cursor, limit: $limit) {
      ... on PartitionBackfills {
        results {
          id
          status
          numPartitions
          timestamp
          partitionNames
          partitionSetName
          partitionSet {
            id
            name
            mode
            pipelineName
            repositoryOrigin {
              id
              repositoryName
              repositoryLocationName
            }
          }
        }
      }
    }
  }
"""

SINGLE_BACKFILL_QUERY = """
  query SingleBackfillQuery($backfillId: String!) {
    partitionBackfillOrError(backfillId: $backfillId) {
      ... on PartitionBackfill {
        hasCancelPermission
        hasResumePermission
        partitionStatuses {
          results {
            id
            partitionName
            runId
            runStatus
          }
        }
      }
    }
  }
"""

ASSET_BACKFILL_DATA_QUERY = """
  query BackfillStatusesByAsset($backfillId: String!) {
    partitionBackfillOrError(backfillId: $backfillId) {
      ... on PartitionBackfill {
        assetBackfillData {
            rootTargetedPartitions {
                partitionKeys
                ranges {
                    start
                    end
                }
            }
        }
        isAssetBackfill
      }
    }
  }
"""

ASSET_BACKFILL_LOGS_QUERY = """
  query BackfillLogsByAsset($backfillId: String!) {
    partitionBackfillOrError(backfillId: $backfillId) {
      ... on PartitionBackfill {
        logEvents {
            events {
                message
                timestamp
                level
            }
            cursor
        }
      }
    }
  }
"""

ASSET_BACKFILL_PREVIEW_QUERY = """
query assetBackfillPreview($params: AssetBackfillPreviewParams!) {
  assetBackfillPreview(params: $params) {
    assetKey {
      path
    }
    partitions {
      partitionKeys
      ranges {
        start
        end
      }
    }
  }
}
"""

BACKFILLS_WITH_FILTERS_QUERY = """
  query BackfillsWithFiltersQuery($filters: BulkActionsFilter) {
    partitionBackfillsOrError(filters: $filters) {
      ... on PartitionBackfills {
        results {
          id
          timestamp
          status
        }
      }
    }
  }
"""


def get_repo() -> RepositoryDefinition:
    partitions_def = StaticPartitionsDefinition(["a", "b", "c", "d", "e", "f", "g"])

    @asset(partitions_def=partitions_def)
    def asset1(): ...

    @asset(partitions_def=partitions_def)
    def asset2(): ...

    @asset()
    def asset3():
        """Non-partitioned asset."""
        ...

    return Definitions(assets=[asset1, asset2, asset3]).get_repository_def()


def get_repo_with_non_partitioned_asset() -> RepositoryDefinition:
    partitions_def = StaticPartitionsDefinition(["a", "b", "c"])

    @asset(partitions_def=partitions_def)
    def asset1(): ...

    @asset
    def asset2(asset1): ...

    return Definitions(assets=[asset1, asset2]).get_repository_def()


def get_repo_with_non_subsettable_multi_asset() -> RepositoryDefinition:
    partitions_def = StaticPartitionsDefinition(["x", "y", "z"])

    @multi_asset(outs={"a": AssetOut(), "b": AssetOut()}, partitions_def=partitions_def)
    def non_subsettable_multi_asset(): ...

    return Definitions(assets=[non_subsettable_multi_asset]).get_repository_def()


def get_repo_with_missing_upstream_partition() -> RepositoryDefinition:
    upstream_partitions_def = StaticPartitionsDefinition(["a", "b"])
    downstream_partitions_def = StaticPartitionsDefinition(["a", "b", "c"])

    @asset(partitions_def=downstream_partitions_def)
    def child(parent): ...

    @asset(partitions_def=upstream_partitions_def)
    def parent(): ...

    return Definitions(assets=[child, parent]).get_repository_def()


def get_repo_with_dynamic_partition() -> RepositoryDefinition:
    dynamic_partitions_def = DynamicPartitionsDefinition(name="dynamic")

    @asset(partitions_def=dynamic_partitions_def)
    def the_asset(): ...

    return Definitions(assets=[the_asset]).get_repository_def()


def get_repo_with_root_assets_different_partitions() -> RepositoryDefinition:
    return Definitions(assets=root_assets_different_partitions_same_downstream).get_repository_def()


def test_launch_asset_backfill_read_only_context():
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        # read-only context fails
        with define_out_of_process_context(
            __file__, "get_repo", instance, read_only=True
        ) as read_only_context:
            assert read_only_context.read_only
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                read_only_context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "b"],
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            assert launch_backfill_result
            assert launch_backfill_result.data

            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "UnauthorizedError"
            )

            launch_backfill_result = execute_dagster_graphql(
                read_only_context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionsByAssets": [
                            {
                                "assetKey": key.to_graphql_input(),
                                "partitions": {"range": {"start": "a", "end": "b"}},
                            }
                            for key in all_asset_keys
                        ]
                    }
                },
            )

            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "UnauthorizedError"
            )

        location_name = main_repo_location_name()

        # context with per-location permissions on the specific location succeeds
        with define_out_of_process_context(
            __file__,
            "get_repo",
            instance,
            read_only=True,
            read_only_locations={location_name: False},  # not read-only in this specific location
        ) as read_only_context:
            assert read_only_context.read_only
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                read_only_context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "b"],
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            assert launch_backfill_result
            assert launch_backfill_result.data

            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "LaunchBackfillSuccess"
            )

            backfill_id = launch_backfill_result.data["launchPartitionBackfill"]["backfillId"]

            single_backfill_result = execute_dagster_graphql(
                read_only_context, SINGLE_BACKFILL_QUERY, variables={"backfillId": backfill_id}
            )

            assert single_backfill_result.data
            assert single_backfill_result.data["partitionBackfillOrError"]["hasCancelPermission"]
            assert single_backfill_result.data["partitionBackfillOrError"]["hasResumePermission"]

            launch_backfill_result = execute_dagster_graphql(
                read_only_context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionsByAssets": [
                            {
                                "assetKey": key.to_graphql_input(),
                                "partitions": {"range": {"start": "a", "end": "b"}},
                            }
                            for key in all_asset_keys
                        ]
                    }
                },
            )

            assert launch_backfill_result
            assert launch_backfill_result.data

            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "LaunchBackfillSuccess"
            )

            # assets that aren't in the asset graph at all fail permissions check
            # because they can't be mapped to a particular code location
            launch_backfill_result = execute_dagster_graphql(
                read_only_context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "b"],
                        "assetSelection": [{"path": ["doesnot", "exist"]}],
                    }
                },
            )
            assert launch_backfill_result
            assert launch_backfill_result.data

            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "UnauthorizedError"
            )

            launch_backfill_result = execute_dagster_graphql(
                read_only_context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionsByAssets": [
                            {
                                "assetKey": {"path": ["doesnot", "exist"]},
                                "partitions": {"range": {"start": "a", "end": "b"}},
                            }
                        ]
                    }
                },
            )

            assert launch_backfill_result
            assert launch_backfill_result.data

            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "UnauthorizedError"
            )


def test_launch_asset_backfill_all_partitions():
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                        "allPartitions": True,
                    }
                },
            )

            _backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            target_subset = asset_backfill_data.target_subset
            assert target_subset.asset_keys == all_asset_keys
            all_partition_keys = {"a", "b", "c", "d", "e", "f", "g"}
            assert (
                target_subset.get_partitions_subset(AssetKey("asset1")).get_partition_keys()
                == all_partition_keys
            )
            assert (
                target_subset.get_partitions_subset(AssetKey("asset2")).get_partition_keys()
                == all_partition_keys
            )


def test_launch_asset_backfill_all_partitions_asset_selection():
    repo = get_repo()

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "assetSelection": [AssetKey("asset2").to_graphql_input()],
                        "allPartitions": True,
                    }
                },
            )

            _backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            target_subset = asset_backfill_data.target_subset
            assert target_subset.asset_keys == {AssetKey("asset2")}
            all_partition_keys = {"a", "b", "c", "d", "e", "f", "g"}
            assert (
                target_subset.get_partitions_subset(AssetKey("asset2")).get_partition_keys()
                == all_partition_keys
            )
            assert AssetKey("asset1") not in target_subset.asset_keys


def test_launch_asset_backfill_partitions_by_asset():
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionsByAssets": [
                            {
                                "assetKey": AssetKey("asset1").to_graphql_input(),
                                "partitions": {
                                    "range": {
                                        "start": "b",
                                        "end": "c",
                                    }
                                },
                            },
                            {
                                "assetKey": AssetKey("asset2").to_graphql_input(),
                            },
                            {
                                "assetKey": AssetKey("asset3").to_graphql_input(),
                            },
                        ],
                    }
                },
            )

            _backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            target_subset = asset_backfill_data.target_subset
            assert target_subset.asset_keys == all_asset_keys
            all_partition_keys = {"a", "b", "c", "d", "e", "f", "g"}
            assert target_subset.get_partitions_subset(AssetKey("asset1")).get_partition_keys() == {
                "b",
                "c",
            }
            assert (
                target_subset.get_partitions_subset(AssetKey("asset2")).get_partition_keys()
                == all_partition_keys
            )
            assert target_subset.non_partitioned_asset_keys == {AssetKey("asset3")}


def test_launch_asset_backfill_partitions_by_asset_multiple_ranges():
    repo = get_repo()

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionsByAssets": [
                            {
                                "assetKey": AssetKey("asset1").to_graphql_input(),
                                "partitions": {
                                    "ranges": [
                                        {"start": "b", "end": "d"},
                                        {"start": "f", "end": "g"},
                                    ]
                                },
                            },
                            {"assetKey": AssetKey("asset2").to_graphql_input()},
                        ],
                    }
                },
            )

            _backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            target_subset = asset_backfill_data.target_subset
            all_partition_keys = {"a", "b", "c", "d", "e", "f", "g"}
            target_partition_keys = {"b", "c", "d", "f", "g"}

            assert (
                target_subset.get_partitions_subset(AssetKey("asset1")).get_partition_keys()
                == target_partition_keys
            )
            assert (
                target_subset.get_partitions_subset(AssetKey("asset2")).get_partition_keys()
                == all_partition_keys
            )


def test_launch_asset_backfill_all_partitions_root_assets_different_partitions():
    repo = get_repo_with_root_assets_different_partitions()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_with_root_assets_different_partitions", instance
        ) as context:
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                        "allPartitions": True,
                    }
                },
            )

            _backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            target_subset = asset_backfill_data.target_subset
            assert target_subset.get_partitions_subset(AssetKey("root1")).get_partition_keys() == {
                "a",
                "b",
            }
            assert target_subset.get_partitions_subset(AssetKey("root2")).get_partition_keys() == {
                "1",
                "2",
            }
            assert target_subset.get_partitions_subset(
                AssetKey("downstream")
            ).get_partition_keys() == {"a", "b"}


def test_launch_asset_backfill():
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "b"],
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            assert asset_backfill_data.target_subset.asset_keys == all_asset_keys

            # on PartitionBackfills
            get_backfills_result = execute_dagster_graphql(
                context, GET_PARTITION_BACKFILLS_QUERY, variables={}
            )
            assert not get_backfills_result.errors
            assert get_backfills_result.data
            backfill_results = get_backfills_result.data["partitionBackfillsOrError"]["results"]
            assert len(backfill_results) == 1
            assert backfill_results[0]["numPartitions"] == 2
            assert backfill_results[0]["id"] == backfill_id
            assert backfill_results[0]["partitionSet"] is None
            assert backfill_results[0]["partitionSetName"] is None
            assert set(backfill_results[0]["partitionNames"]) == {"a", "b"}

            # on PartitionBackfill
            single_backfill_result = execute_dagster_graphql(
                context, SINGLE_BACKFILL_QUERY, variables={"backfillId": backfill_id}
            )
            assert not single_backfill_result.errors
            assert single_backfill_result.data
            assert (
                single_backfill_result.data["partitionBackfillOrError"]["partitionStatuses"] is None
            )


def test_launch_asset_backfill_with_nonexistent_partition_key():
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "nonexistent1", "nonexistent2"],
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "PartitionKeysNotFoundError"
            )
            assert (
                "Partition keys `['nonexistent1', 'nonexistent2']` could not be found"
                in launch_backfill_result.data["launchPartitionBackfill"]["message"]
            )


def test_launch_asset_backfill_with_nonexistent_dynamic_partition_key():
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        instance.add_dynamic_partitions("dynamic_partition", ["a", "b", "c"])
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["b", "nonexistent1", "nonexistent2"],
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "PartitionKeysNotFoundError"
            )
            assert (
                "Partition keys `['nonexistent1', 'nonexistent2']` could not be found"
                in launch_backfill_result.data["launchPartitionBackfill"]["message"]
            )


def test_launch_asset_backfill_with_nonexistent_upstream_partition_key():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_with_missing_upstream_partition", instance
        ) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "b", "c"],
                        "assetSelection": [{"path": ["child"]}],
                    }
                },
            )
            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "PythonError"
            )
            assert (
                "depends on non-existent partitions: EntitySubset<AssetKey(['parent'])>(DefaultPartitionsSubset(subset={'c'}"
                in launch_backfill_result.data["launchPartitionBackfill"]["message"]
            )


def test_launch_asset_backfill_with_non_subsettable_multi_asset():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_with_non_subsettable_multi_asset", instance
        ) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["x", "y"],
                        "assetSelection": [{"path": ["a"]}],
                    }
                },
            )
            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "PythonError"
            )
            assert (
                "Backfill must include every key in a non-subsettable multi-asset, included a but not b"
                in launch_backfill_result.data["launchPartitionBackfill"]["message"]
            )

            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionsByAssets": [
                            {
                                "assetKey": {"path": ["a"]},
                                "partitions": {
                                    "range": {
                                        "start": "x",
                                        "end": "x",
                                    }
                                },
                            },
                            {
                                "assetKey": {"path": ["b"]},
                                "partitions": {
                                    "range": {
                                        "start": "x",
                                        "end": "y",
                                    }
                                },
                            },
                        ],
                    }
                },
            )
            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "PythonError"
            )
            assert (
                "Targeted subset must match between every key in a non-subsettable multi-asset, but does not match between"
                in launch_backfill_result.data["launchPartitionBackfill"]["message"]
            )

            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["x", "y"],
                        "assetSelection": [{"path": ["a"]}, {"path": ["b"]}],
                    }
                },
            )
            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "LaunchBackfillSuccess"
            )


def test_remove_partitions_defs_after_backfill_backcompat():
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "b"],
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            assert asset_backfill_data.target_subset.asset_keys == all_asset_keys

        # Replace the asset backfill data with the backcompat serialization
        backfill = instance.get_backfills()[0]
        backcompat_backfill = backfill._replace(
            asset_backfill_data=None,
            serialized_asset_backfill_data=backfill.asset_backfill_data.serialize(  # pyright: ignore[reportOptionalMemberAccess]
                instance, asset_graph=repo.asset_graph
            ),
        )

        with mock.patch(
            "dagster._core.instance.DagsterInstance.get_backfills",
            return_value=[backcompat_backfill],
        ):
            # When the partitions defs are unchanged, the backfill data can be fetched
            with define_out_of_process_context(__file__, "get_repo", instance) as context:
                get_backfills_result = execute_dagster_graphql(
                    context, GET_PARTITION_BACKFILLS_QUERY, variables={}
                )
                assert not get_backfills_result.errors
                assert get_backfills_result.data

                backfill_results = get_backfills_result.data["partitionBackfillsOrError"]["results"]
                assert len(backfill_results) == 1
                assert backfill_results[0]["numPartitions"] == 2
                assert backfill_results[0]["id"] == backfill_id
                assert set(backfill_results[0]["partitionNames"]) == {"a", "b"}

            # When the partitions defs are changed, the backfill data cannot be fetched
            with define_out_of_process_context(
                __file__, "get_repo_with_non_partitioned_asset", instance
            ) as context:
                get_backfills_result = execute_dagster_graphql(
                    context, GET_PARTITION_BACKFILLS_QUERY, variables={}
                )
                assert not get_backfills_result.errors
                assert get_backfills_result.data

                backfill_results = get_backfills_result.data["partitionBackfillsOrError"]["results"]
                assert len(backfill_results) == 1
                assert backfill_results[0]["numPartitions"] == 0
                assert backfill_results[0]["id"] == backfill_id
                assert set(backfill_results[0]["partitionNames"]) == set()


def test_remove_partitions_defs_after_backfill():
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "b"],
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            assert asset_backfill_data.target_subset.asset_keys == all_asset_keys

        with define_out_of_process_context(
            __file__, "get_repo_with_non_partitioned_asset", instance
        ) as context:
            # on PartitionBackfills
            get_backfills_result = execute_dagster_graphql(
                context, GET_PARTITION_BACKFILLS_QUERY, variables={}
            )
            assert not get_backfills_result.errors
            assert get_backfills_result.data
            backfill_results = get_backfills_result.data["partitionBackfillsOrError"]["results"]
            assert len(backfill_results) == 1
            assert backfill_results[0]["numPartitions"] == 2
            assert backfill_results[0]["id"] == backfill_id
            assert backfill_results[0]["partitionSet"] is None
            assert backfill_results[0]["partitionSetName"] is None
            assert set(backfill_results[0]["partitionNames"]) == {"a", "b"}

            # on PartitionBackfill
            single_backfill_result = execute_dagster_graphql(
                context, SINGLE_BACKFILL_QUERY, variables={"backfillId": backfill_id}
            )
            assert not single_backfill_result.errors
            assert single_backfill_result.data
            assert (
                single_backfill_result.data["partitionBackfillOrError"]["partitionStatuses"] is None
            )


def test_launch_asset_backfill_with_non_partitioned_asset():
    repo = get_repo_with_non_partitioned_asset()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_with_non_partitioned_asset", instance
        ) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "b"],
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            _backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            target_subset = asset_backfill_data.target_subset
            assert target_subset.asset_keys == all_asset_keys
            assert target_subset.get_partitions_subset(AssetKey("asset1")).get_partition_keys() == {
                "a",
                "b",
            }
            assert AssetKey("asset2") in target_subset.non_partitioned_asset_keys
            assert AssetKey("asset2") not in target_subset.partitions_subsets_by_asset_key


def get_daily_hourly_repo() -> RepositoryDefinition:
    @asset(partitions_def=HourlyPartitionsDefinition(start_date="2020-01-01-00:00"))
    def hourly(): ...

    @asset(partitions_def=DailyPartitionsDefinition(start_date="2020-01-01"))
    def daily(hourly): ...

    return Definitions(assets=[hourly, daily]).get_repository_def()


def test_launch_asset_backfill_with_upstream_anchor_asset():
    repo = get_daily_hourly_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    hourly_partitions = ["2020-01-02-23:00", "2020-01-02-22:00", "2020-01-03-00:00"]

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_daily_hourly_repo", instance) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": hourly_partitions,
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            target_subset = asset_backfill_data.target_subset
            asset_graph = repo.asset_graph
            assert target_subset == AssetGraphSubset(
                partitions_subsets_by_asset_key={
                    AssetKey("hourly"): asset_graph.get(
                        AssetKey("hourly")
                    ).partitions_def.subset_with_partition_keys(hourly_partitions),  # pyright: ignore[reportOptionalMemberAccess]
                    AssetKey("daily"): asset_graph.get(
                        AssetKey("daily")
                    ).partitions_def.subset_with_partition_keys(["2020-01-02", "2020-01-03"]),  # pyright: ignore[reportOptionalMemberAccess]
                },
            )

            # on PartitionBackfills
            get_backfills_result = execute_dagster_graphql(
                context, GET_PARTITION_BACKFILLS_QUERY, variables={}
            )
            assert not get_backfills_result.errors
            assert get_backfills_result.data
            backfill_results = get_backfills_result.data["partitionBackfillsOrError"]["results"]
            assert len(backfill_results) == 1
            assert backfill_results[0]["numPartitions"] is None
            assert backfill_results[0]["id"] == backfill_id
            assert backfill_results[0]["partitionSet"] is None
            assert backfill_results[0]["partitionSetName"] is None
            assert backfill_results[0]["partitionNames"] is None


def get_daily_two_hourly_repo() -> RepositoryDefinition:
    @asset(partitions_def=HourlyPartitionsDefinition(start_date="2020-01-01-00:00"))
    def hourly1(): ...

    @asset(partitions_def=HourlyPartitionsDefinition(start_date="2020-01-01-00:00"))
    def hourly2(): ...

    @asset(partitions_def=DailyPartitionsDefinition(start_date="2020-01-01"))
    def daily(hourly1, hourly2): ...

    return Definitions(assets=[hourly1, hourly2, daily]).get_repository_def()


def test_launch_asset_backfill_with_two_anchor_assets():
    repo = get_daily_two_hourly_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    hourly_partitions = ["2020-01-02-23:00", "2020-01-02-22:00", "2020-01-03-00:00"]

    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_daily_two_hourly_repo", instance
        ) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": hourly_partitions,
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            _backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            target_subset = asset_backfill_data.target_subset
            asset_graph = repo.asset_graph
            assert target_subset == AssetGraphSubset(
                partitions_subsets_by_asset_key={
                    AssetKey("hourly1"): asset_graph.get(
                        AssetKey("hourly1")
                    ).partitions_def.subset_with_partition_keys(hourly_partitions),  # pyright: ignore[reportOptionalMemberAccess]
                    AssetKey("hourly2"): asset_graph.get(
                        AssetKey("hourly2")
                    ).partitions_def.subset_with_partition_keys(hourly_partitions),  # pyright: ignore[reportOptionalMemberAccess]
                    AssetKey("daily"): asset_graph.get(
                        AssetKey("daily")
                    ).partitions_def.subset_with_partition_keys(["2020-01-02", "2020-01-03"]),  # pyright: ignore[reportOptionalMemberAccess]
                },
            )


def get_daily_hourly_non_partitioned_repo() -> RepositoryDefinition:
    @asset(partitions_def=HourlyPartitionsDefinition(start_date="2020-01-01-00:00"))
    def hourly(): ...

    @asset(partitions_def=DailyPartitionsDefinition(start_date="2020-01-01"))
    def daily(hourly): ...

    @asset
    def non_partitioned(hourly): ...

    return Definitions(assets=[hourly, daily, non_partitioned]).get_repository_def()


def test_launch_asset_backfill_with_upstream_anchor_asset_and_non_partitioned_asset():
    repo = get_daily_hourly_non_partitioned_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    hourly_partitions = ["2020-01-02-23:00", "2020-01-02-22:00", "2020-01-03-00:00"]

    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_daily_hourly_non_partitioned_repo", instance
        ) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": hourly_partitions,
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            target_subset = asset_backfill_data.target_subset
            asset_graph = repo.asset_graph
            assert target_subset == AssetGraphSubset(
                non_partitioned_asset_keys={AssetKey("non_partitioned")},
                partitions_subsets_by_asset_key={
                    AssetKey("hourly"): (
                        asset_graph.get(AssetKey("hourly"))
                        .partitions_def.empty_subset()  # pyright: ignore[reportOptionalMemberAccess]
                        .with_partition_keys(hourly_partitions)
                    ),
                    AssetKey("daily"): (
                        asset_graph.get(AssetKey("daily"))
                        .partitions_def.empty_subset()  # pyright: ignore[reportOptionalMemberAccess]
                        .with_partition_keys(["2020-01-02", "2020-01-03"])
                    ),
                },
            )

            asset_backfill_data_result = execute_dagster_graphql(
                context, ASSET_BACKFILL_DATA_QUERY, variables={"backfillId": backfill_id}
            )
            assert asset_backfill_data_result.data
            assert (
                asset_backfill_data_result.data["partitionBackfillOrError"]["isAssetBackfill"]
                is True
            )
            targeted_ranges = asset_backfill_data_result.data["partitionBackfillOrError"][
                "assetBackfillData"
            ]["rootTargetedPartitions"]["ranges"]
            assert len(targeted_ranges) == 1
            assert targeted_ranges[0]["start"] == "2020-01-02-22:00"
            assert targeted_ranges[0]["end"] == "2020-01-03-00:00"


def test_asset_backfill_non_existant_asset_keys():
    repo = get_daily_hourly_non_partitioned_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    hourly_partitions = ["2020-01-02-23:00", "2020-01-02-22:00", "2020-01-03-00:00"]

    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_daily_hourly_non_partitioned_repo", instance
        ) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": hourly_partitions,
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            backfill_id, _asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )

            asset_backfill_data_result = execute_dagster_graphql(
                context, ASSET_BACKFILL_DATA_QUERY, variables={"backfillId": backfill_id}
            )
            assert asset_backfill_data_result.data
            assert (
                asset_backfill_data_result.data["partitionBackfillOrError"]["isAssetBackfill"]
                is True
            )
            targeted_ranges = asset_backfill_data_result.data["partitionBackfillOrError"][
                "assetBackfillData"
            ]["rootTargetedPartitions"]["ranges"]
            assert len(targeted_ranges) == 1
            assert targeted_ranges[0]["start"] == "2020-01-02-22:00"
            assert targeted_ranges[0]["end"] == "2020-01-03-00:00"

        # Load the backfill data in a different context where the targeted asset keys no longer exist
        with define_out_of_process_context(__file__, "get_repo", instance) as other_context:
            asset_backfill_data_result = execute_dagster_graphql(
                other_context, ASSET_BACKFILL_DATA_QUERY, variables={"backfillId": backfill_id}
            )
            assert asset_backfill_data_result.data
            assert (
                asset_backfill_data_result.data["partitionBackfillOrError"]["isAssetBackfill"]
                is True
            )
            assert (
                asset_backfill_data_result.data["partitionBackfillOrError"]["assetBackfillData"][
                    "rootTargetedPartitions"
                ]
                is None
            )


def test_asset_backfill_status_only_unpartitioned_assets():
    repo = get_daily_hourly_non_partitioned_repo()

    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_daily_hourly_non_partitioned_repo", instance
        ) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "allPartitions": True,
                        "assetSelection": [AssetKey("non_partitioned").to_graphql_input()],
                    }
                },
            )
            backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            target_subset = asset_backfill_data.target_subset

            assert target_subset == AssetGraphSubset(
                non_partitioned_asset_keys={AssetKey("non_partitioned")},
            )

            asset_backfill_data_result = execute_dagster_graphql(
                context, ASSET_BACKFILL_DATA_QUERY, variables={"backfillId": backfill_id}
            )
            assert asset_backfill_data_result.data
            assert (
                asset_backfill_data_result.data["partitionBackfillOrError"]["isAssetBackfill"]
                is True
            )
            assert (
                asset_backfill_data_result.data["partitionBackfillOrError"]["assetBackfillData"][
                    "rootTargetedPartitions"
                ]
                is None
            )


def test_asset_backfill_preview_time_partitioned():
    repo = get_daily_hourly_non_partitioned_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    hourly_partitions = ["2020-01-02-23:00", "2020-01-02-22:00", "2020-01-03-00:00"]

    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_daily_hourly_non_partitioned_repo", instance
        ) as context:
            backfill_preview_result = execute_dagster_graphql(
                context,
                ASSET_BACKFILL_PREVIEW_QUERY,
                variables={
                    "params": {
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                        "partitionNames": hourly_partitions,
                    }
                },
            )

            target_asset_partitions = backfill_preview_result.data["assetBackfillPreview"]
            assert len(target_asset_partitions) == 3

            # Assert toposorted
            assert target_asset_partitions[0]["assetKey"] == {"path": ["hourly"]}
            assert target_asset_partitions[0]["partitions"]["ranges"] == [
                {"start": "2020-01-02-22:00", "end": "2020-01-03-00:00"}
            ]
            assert target_asset_partitions[0]["partitions"]["partitionKeys"] is None

            assert target_asset_partitions[1]["assetKey"] == {"path": ["daily"]}
            assert target_asset_partitions[1]["partitions"]["ranges"] == [
                {"start": "2020-01-02", "end": "2020-01-03"}
            ]
            assert target_asset_partitions[1]["partitions"]["partitionKeys"] is None

            assert target_asset_partitions[2]["assetKey"] == {"path": ["non_partitioned"]}
            assert target_asset_partitions[2]["partitions"] is None


def test_asset_backfill_preview_static_partitioned():
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys
    partition_keys = ["a", "b"]

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            backfill_preview_result = execute_dagster_graphql(
                context,
                ASSET_BACKFILL_PREVIEW_QUERY,
                variables={
                    "params": {
                        "partitionNames": partition_keys,
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )

            target_asset_partitions = backfill_preview_result.data["assetBackfillPreview"]
            assert len(target_asset_partitions) == 3

            # Assert toposorted
            assert target_asset_partitions[0]["assetKey"] == {"path": ["asset1"]}
            assert target_asset_partitions[0]["partitions"]["ranges"] is None
            assert set(target_asset_partitions[0]["partitions"]["partitionKeys"]) == set(
                partition_keys
            )

            assert target_asset_partitions[1]["assetKey"] == {"path": ["asset2"]}
            assert target_asset_partitions[1]["partitions"]["ranges"] is None
            assert set(target_asset_partitions[1]["partitions"]["partitionKeys"]) == set(
                partition_keys
            )

            assert target_asset_partitions[2]["assetKey"] == {"path": ["asset3"]}
            assert target_asset_partitions[2]["partitions"] is None


def test_asset_backfill_error_raised_upon_invalid_params_provided():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_daily_hourly_non_partitioned_repo", instance
        ) as context:
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionsByAssets": [
                            {
                                "assetKey": {"path": ["hourly"]},
                                "partitions": {
                                    "range": {
                                        "start": "2024-01-01-00:00",
                                        "end": "2024-01-01-01:00",
                                    }
                                },
                            }
                        ],
                        "assetSelection": [{"path": ["hourly"]}],
                    }
                },
            )
            assert launch_backfill_result.data
            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "PythonError"
            )
            assert (
                "partitions_by_assets cannot be used together with asset_selection, selector, or partitionNames"
                in launch_backfill_result.data["launchPartitionBackfill"]["message"]
            )


def test_asset_backfill_error_raised_upon_conflicting_range_params_provided():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_daily_hourly_non_partitioned_repo", instance
        ) as context:
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionsByAssets": [
                            {
                                "assetKey": {"path": ["hourly"]},
                                "partitions": {
                                    "range": {
                                        "start": "2024-01-01-00:00",
                                        "end": "2024-01-01-01:00",
                                    },
                                    "ranges": [
                                        {
                                            "start": "2024-01-01-00:00",
                                            "end": "2024-01-01-01:00",
                                        }
                                    ],
                                },
                            }
                        ],
                    }
                },
            )
            assert launch_backfill_result.data
            assert (
                launch_backfill_result.data["launchPartitionBackfill"]["__typename"]
                == "PythonError"
            )
            assert (
                "Received partitionsSelector with values for both 'range' and 'ranges'. Only one should be provided."
                in launch_backfill_result.data["launchPartitionBackfill"]["message"]
            )


def _get_backfill_data(
    launch_backfill_result: GqlResult,
    instance: DagsterInstance,
    repo,
) -> tuple[str, AssetBackfillData]:
    assert launch_backfill_result
    assert launch_backfill_result.data
    assert "backfillId" in launch_backfill_result.data["launchPartitionBackfill"], (
        _get_error_message(launch_backfill_result)
    )

    backfill_id = launch_backfill_result.data["launchPartitionBackfill"]["backfillId"]

    backfills = instance.get_backfills()
    assert len(backfills) == 1
    backfill = backfills[0]
    assert backfill.backfill_id == backfill_id
    assert backfill.asset_backfill_data

    return backfill_id, backfill.asset_backfill_data


def _get_error_message(launch_backfill_result: GqlResult) -> str | None:
    return (
        (
            "".join(launch_backfill_result.data["launchPartitionBackfill"]["stack"])
            + launch_backfill_result.data["launchPartitionBackfill"]["message"]
        )
        if "message" in launch_backfill_result.data["launchPartitionBackfill"]
        else None
    )


def test_backfill_logs():
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        # need to override this method on the instance since it defaults ot False in OSS. When we enable this
        # feature in OSS we can remove this override
        def override_backfill_storage_setting(self):
            return True

        instance.backfill_log_storage_enabled = override_backfill_storage_setting.__get__(
            instance, DagsterInstance
        )
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            # launchPartitionBackfill
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "b"],
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                    }
                },
            )
            backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )
            assert asset_backfill_data.target_subset.asset_keys == all_asset_keys

            list(
                execute_backfill_iteration(
                    context.process_context, get_default_daemon_logger("BackfillDaemon")
                )
            )

            backfill_logs = execute_dagster_graphql(
                context,
                ASSET_BACKFILL_LOGS_QUERY,
                variables={
                    "backfillId": backfill_id,
                },
            )

            assert len(backfill_logs.data["partitionBackfillOrError"]["logEvents"]["events"]) > 0


def test_get_backfills_with_filters():
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            # launchPartitionBackfill
            all_backfills = []
            for _ in range(5):
                launch_backfill_result = execute_dagster_graphql(
                    context,
                    LAUNCH_PARTITION_BACKFILL_MUTATION,
                    variables={
                        "backfillParams": {
                            "partitionNames": ["a", "b"],
                            "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                        }
                    },
                )
                assert "backfillId" in launch_backfill_result.data["launchPartitionBackfill"], (
                    _get_error_message(launch_backfill_result)
                )

                backfill_id = launch_backfill_result.data["launchPartitionBackfill"]["backfillId"]
                backfill = instance.get_backfill(backfill_id)
                all_backfills.append(backfill)

            get_backfills_result = execute_dagster_graphql(
                context,
                BACKFILLS_WITH_FILTERS_QUERY,
                variables={"filters": {"createdBefore": all_backfills[3].backfill_timestamp}},
            )
            assert not get_backfills_result.errors
            assert get_backfills_result.data
            backfill_results = get_backfills_result.data["partitionBackfillsOrError"]["results"]

            for result in backfill_results:
                assert result["timestamp"] < all_backfills[3].backfill_timestamp

            get_backfills_result = execute_dagster_graphql(
                context,
                BACKFILLS_WITH_FILTERS_QUERY,
                variables={"filters": {"createdAfter": all_backfills[3].backfill_timestamp}},
            )
            assert not get_backfills_result.errors
            assert get_backfills_result.data
            backfill_results = get_backfills_result.data["partitionBackfillsOrError"]["results"]

            for result in backfill_results:
                assert result["timestamp"] > all_backfills[3].backfill_timestamp

            get_backfills_result = execute_dagster_graphql(
                context,
                BACKFILLS_WITH_FILTERS_QUERY,
                variables={"filters": {"statuses": ["REQUESTED"]}},
            )


def test_launch_asset_backfill_with_run_config_data():
    """Test that asset backfill can be launched with runConfigData."""
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            # Test with runConfigData containing various configuration options
            run_config_data = {
                "ops": {
                    "asset1": {"config": {"some_param": "test_value", "another_param": 42}},
                    "asset2": {"config": {"nested_config": {"key": "value"}}},
                },
                "resources": {"some_resource": {"config": {"resource_param": "resource_value"}}},
            }

            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "b"],
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                        "runConfigData": run_config_data,
                    }
                },
            )

            backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )

            # Verify the backfill was created successfully
            assert asset_backfill_data.target_subset.asset_keys == all_asset_keys

            # Verify the runConfigData was stored correctly
            backfill = instance.get_backfill(backfill_id)
            assert backfill is not None
            assert backfill.run_config == run_config_data

            # Test that the backfill can be queried and shows the correct configuration
            single_backfill_result = execute_dagster_graphql(
                context, SINGLE_BACKFILL_QUERY, variables={"backfillId": backfill_id}
            )
            assert not single_backfill_result.errors
            assert single_backfill_result.data


def test_launch_asset_backfill_with_empty_run_config_data():
    """Test that asset backfill can be launched with empty runConfigData."""
    repo = get_repo()
    all_asset_keys = repo.asset_graph.materializable_asset_keys

    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            # Test with empty runConfigData
            launch_backfill_result = execute_dagster_graphql(
                context,
                LAUNCH_PARTITION_BACKFILL_MUTATION,
                variables={
                    "backfillParams": {
                        "partitionNames": ["a", "b"],
                        "assetSelection": [key.to_graphql_input() for key in all_asset_keys],
                        "runConfigData": {},
                    }
                },
            )

            backfill_id, asset_backfill_data = _get_backfill_data(
                launch_backfill_result, instance, repo
            )

            # Verify the backfill was created successfully
            assert asset_backfill_data.target_subset.asset_keys == all_asset_keys

            # Verify the runConfigData was stored correctly
            backfill = instance.get_backfill(backfill_id)
            assert backfill is not None
            assert backfill.run_config == {}



import graphene
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AutomationConditionSnapshot,
    get_expanded_label,
)

from dagster_graphql.schema.util import non_null_list


class GrapheneAutomationCondition(graphene.ObjectType):
    label = graphene.Field(graphene.String)
    expandedLabel = non_null_list(graphene.String)

    class Meta:
        name = "AutomationCondition"

    def __init__(self, snapshot: AutomationConditionSnapshot):
        super().__init__(
            label=snapshot.node_snapshot.label,
            expandedLabel=get_expanded_label(snapshot),
        )



from dagster._core.workspace.context import WorkspaceRequestContext
from dagster_graphql.test.utils import execute_dagster_graphql, infer_repository_selector

INPUT_OUTPUT_DEFINITIONS_QUERY = """
    query InputOutputDefinitionsQuery($repositorySelector: RepositorySelector!) {
        repositoryOrError(repositorySelector: $repositorySelector) {
           ... on Repository {
                usedSolid(name: "op_with_input_output_metadata") {
                    __typename
                    definition {
                        inputDefinitions {
                            metadataEntries {
                                label
                            }
                        }
                        outputDefinitions {
                            metadataEntries {
                                label
                            }
                        }
                    }
                }
            }
        }
    }
"""


def test_query_inputs_outputs(graphql_context: WorkspaceRequestContext, snapshot):
    selector = infer_repository_selector(graphql_context)
    result = execute_dagster_graphql(
        graphql_context, INPUT_OUTPUT_DEFINITIONS_QUERY, variables={"repositorySelector": selector}
    )
    assert result.data
    snapshot.assert_match(result.data)



from unittest import mock

import pytest
from dagster._core.workspace.context import WorkspaceRequestContext
from dagster_graphql.implementation.fetch_pipelines import _get_job_snapshot_from_instance
from dagster_graphql.implementation.utils import UserFacingGraphQLError
from dagster_graphql.test.utils import (
    execute_dagster_graphql,
    main_repo_location_name,
    main_repo_name,
)
from dagster_shared.seven import json

from dagster_graphql_tests.graphql.repo import noop_job, owned_job

SNAPSHOT_OR_ERROR_QUERY_BY_SNAPSHOT_ID = """
query PipelineSnapshotQueryBySnapshotID($snapshotId: String!) {
    pipelineSnapshotOrError(snapshotId: $snapshotId) {
        __typename
        ... on PipelineSnapshot {
            name
            pipelineSnapshotId
            description
            dagsterTypes { key }
            solids { name }
            modes { name }
            solidHandles { handleID }
            tags { key value }
            runTags { key value }
            owners {
                ... on UserDefinitionOwner {
                    email
                }
                ... on TeamDefinitionOwner {
                    team
                }
            }
        }
        ... on PipelineSnapshotNotFoundError {
            snapshotId
        }
    }
}
"""

SNAPSHOT_OR_ERROR_QUERY_BY_PIPELINE_NAME = """
query PipelineSnapshotQueryByActivePipelineName($activePipelineSelector: PipelineSelector!) {
    pipelineSnapshotOrError(activePipelineSelector: $activePipelineSelector) {
        __typename
        ... on PipelineSnapshot {
            name
            pipelineSnapshotId
            description
            dagsterTypes { key }
            solids { name }
            modes { name }
            solidHandles { handleID }
            tags { key value }
            runTags { key value }
            externalJobSource
        }
        ... on PipelineSnapshotNotFoundError {
            snapshotId
        }
        ... on PythonError {
            message
            stack
        }
    }
}
"""


# makes snapshot tests much easier to debug
def pretty_dump(data) -> str:
    return json.dumps(data, indent=2, separators=(",", ": "))


def test_fetch_snapshot_or_error_by_snapshot_id_success(
    graphql_context: WorkspaceRequestContext, snapshot
):
    instance = graphql_context.instance
    result = noop_job.execute_in_process(instance=instance)
    assert result.success
    run = instance.get_run_by_id(result.run_id)
    assert run and run.job_snapshot_id

    result = execute_dagster_graphql(
        graphql_context,
        SNAPSHOT_OR_ERROR_QUERY_BY_SNAPSHOT_ID,
        {"snapshotId": run.job_snapshot_id},
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineSnapshotOrError"]["__typename"] == "PipelineSnapshot"

    snapshot.assert_match(pretty_dump(result.data))


def test_job_owners(graphql_context: WorkspaceRequestContext):
    instance = graphql_context.instance
    result = owned_job.execute_in_process(instance=instance)
    assert result.success
    run = instance.get_run_by_id(result.run_id)
    assert run and run.job_snapshot_id

    result = execute_dagster_graphql(
        graphql_context,
        SNAPSHOT_OR_ERROR_QUERY_BY_SNAPSHOT_ID,
        {"snapshotId": run.job_snapshot_id},
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineSnapshotOrError"]["__typename"] == "PipelineSnapshot"

    owners = result.data["pipelineSnapshotOrError"]["owners"]
    assert owners is not None
    assert len(owners) == 2

    # Check the user owner
    user_owner = None
    team_owner = None
    for owner in owners:
        if owner.get("email"):
            user_owner = owner
        elif owner.get("team"):
            team_owner = owner

    assert user_owner is not None
    assert user_owner["email"] == "test@elementl.com"
    assert team_owner is not None
    assert team_owner["team"] == "foo"


def test_fetch_snapshot_or_error_by_snapshot_id_snapshot_not_found(
    graphql_context: WorkspaceRequestContext, snapshot
):
    result = execute_dagster_graphql(
        graphql_context,
        SNAPSHOT_OR_ERROR_QUERY_BY_SNAPSHOT_ID,
        {"snapshotId": "notthere"},
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineSnapshotOrError"]["__typename"] == "PipelineSnapshotNotFoundError"
    assert result.data["pipelineSnapshotOrError"]["snapshotId"] == "notthere"
    snapshot.assert_match(pretty_dump(result.data))


def test_fetch_snapshot_or_error_by_active_pipeline_name_success(
    graphql_context: WorkspaceRequestContext, snapshot
):
    result = execute_dagster_graphql(
        graphql_context,
        SNAPSHOT_OR_ERROR_QUERY_BY_PIPELINE_NAME,
        {
            "activePipelineSelector": {
                "pipelineName": "tagged_job",
                "repositoryName": main_repo_name(),
                "repositoryLocationName": main_repo_location_name(),
            }
        },
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineSnapshotOrError"]["__typename"] == "PipelineSnapshot"
    assert result.data["pipelineSnapshotOrError"]["name"] == "tagged_job"
    assert result.data["pipelineSnapshotOrError"]["tags"] == [{"key": "foo", "value": "bar"}]
    assert result.data["pipelineSnapshotOrError"]["runTags"] == [{"key": "baz", "value": "quux"}]
    assert result.data["pipelineSnapshotOrError"]["externalJobSource"] is None

    snapshot.assert_match(pretty_dump(result.data))


def test_fetch_snapshot_or_error_by_active_pipeline_name_not_found(
    graphql_context: WorkspaceRequestContext, snapshot
):
    result = execute_dagster_graphql(
        graphql_context,
        SNAPSHOT_OR_ERROR_QUERY_BY_PIPELINE_NAME,
        {
            "activePipelineSelector": {
                "pipelineName": "jkdjfkdj",
                "repositoryName": main_repo_name(),
                "repositoryLocationName": main_repo_location_name(),
            }
        },
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineSnapshotOrError"]["__typename"] == "PipelineNotFoundError"

    snapshot.assert_match(pretty_dump(result.data))


def test_external_job_retrieval(graphql_context: WorkspaceRequestContext, snapshot):
    result = execute_dagster_graphql(
        graphql_context,
        SNAPSHOT_OR_ERROR_QUERY_BY_PIPELINE_NAME,
        {
            "activePipelineSelector": {
                "pipelineName": "some_external_job",
                "repositoryName": main_repo_name(),
                "repositoryLocationName": main_repo_location_name(),
            }
        },
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineSnapshotOrError"]["__typename"] == "PipelineSnapshot"
    assert result.data["pipelineSnapshotOrError"]["externalJobSource"] == "airflow"

    snapshot.assert_match(pretty_dump(result.data))


def test_temporary_error_or_deletion_after_instance_check():
    instance = mock.MagicMock()

    instance.has_historical_job.return_value = True
    instance.get_historical_job.return_value = None

    with pytest.raises(UserFacingGraphQLError):
        _get_job_snapshot_from_instance(instance, "kjdkfjd")


def test_fetch_snapshot_or_error_by_snap_and_selector(
    graphql_context: WorkspaceRequestContext, snapshot
):
    # falls back to selector on snap id miss
    result = execute_dagster_graphql(
        graphql_context,
        SNAPSHOT_OR_ERROR_QUERY_BY_PIPELINE_NAME,
        {
            "snapshotId": "notthere",
            "activePipelineSelector": {
                "pipelineName": "csv_hello_world",
                "repositoryName": main_repo_name(),
                "repositoryLocationName": main_repo_location_name(),
            },
        },
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineSnapshotOrError"]["__typename"] == "PipelineSnapshot"
    assert result.data["pipelineSnapshotOrError"]["name"] == "csv_hello_world"

    instance = graphql_context.instance
    result = noop_job.execute_in_process(instance=instance)
    assert result.success
    run = instance.get_run_by_id(result.run_id)
    assert run and run.job_snapshot_id

    # valid snap id but bad selector works (snap preferred)
    result = execute_dagster_graphql(
        graphql_context,
        SNAPSHOT_OR_ERROR_QUERY_BY_SNAPSHOT_ID,
        {
            "snapshotId": run.job_snapshot_id,
            "activePipelineSelector": {
                "pipelineName": "does_not_exist",
                "repositoryName": main_repo_name(),
                "repositoryLocationName": main_repo_location_name(),
            },
        },
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineSnapshotOrError"]["__typename"] == "PipelineSnapshot"
